@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1996-8 Michael R. Elkins <me@cs.hmc.edu>
+ * Copyright (C) 1996-2000 Michael R. Elkins <me@cs.hmc.edu>
  * 
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -13,7 +13,7 @@
  * 
  *     You should have received a copy of the GNU General Public License
  *     along with this program; if not, write to the Free Software
- *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
  */ 
 
 #include "mutt.h"
@@ -23,6 +23,10 @@
 #include "mapping.h"
 #include "sort.h"
 #include "mailbox.h"
+#include "browser.h"
+#ifdef USE_IMAP
+#include "imap.h"
+#endif
 
 #include <stdlib.h>
 #include <dirent.h>
@@ -30,24 +34,7 @@
 #include <ctype.h>
 #include <unistd.h>
 #include <sys/stat.h>
-#include <pwd.h>
-#include <grp.h>
-
-struct folder_file
-{
-  mode_t mode;
-  time_t mtime;
-  off_t size;
-  char *name;
-  char *desc;
-};
-
-struct browser_state
-{
-  struct folder_file *entry;
-  short entrylen; /* number of real entries */
-  short entrymax;  /* max entry */
-};
+#include <errno.h>
 
 static struct mapping_t FolderHelp[] = {
   { N_("Exit"),  OP_EXIT },
@@ -59,12 +46,12 @@ static struct mapping_t FolderHelp[] = {
 
 typedef struct folder_t
 {
-  const char *name;
-  const struct stat *f;
-  int new;
+  struct folder_file *ff;
+  int num;
 } FOLDER;
 
 static char LastDir[_POSIX_PATH_MAX] = "";
+static char LastDirBackup[_POSIX_PATH_MAX] = "";
 
 /* Frees up the memory allocated for the local-global variables.  */
 static void destroy_state (struct browser_state *state)
@@ -75,7 +62,11 @@ static void destroy_state (struct browser_state *state)
   {
     safe_free ((void **) &((state->entry)[c].name));
     safe_free ((void **) &((state->entry)[c].desc));
+    safe_free ((void **) &((state->entry)[c].st));
   }
+#ifdef USE_IMAP
+  safe_free ((void **) &state->folder);
+#endif
   safe_free ((void **) &state->entry);
 }
 
@@ -131,14 +122,17 @@ static void browser_sort (struct browser_state *state)
   qsort (state->entry, state->entrylen, sizeof (struct folder_file), f);
 }
 
-static int link_is_dir (const char *path)
+static int link_is_dir (const char *folder, const char *path)
 {
   struct stat st;
-
-  if (stat (path, &st) == 0)
+  char fullpath[_POSIX_PATH_MAX];
+  
+  snprintf (fullpath, sizeof (fullpath), "%s/%s", folder, path);
+  
+  if (stat (fullpath, &st) == 0)
     return (S_ISDIR (st.st_mode));
   else
-    return (-1);
+    return 0;
 }
 
 static const char *
@@ -152,166 +146,189 @@ folder_format_str (char *dest, size_t destlen, char op, const char *src,
   FOLDER *folder = (FOLDER *) data;
   struct passwd *pw;
   struct group *gr;
+  int optional = (flags & M_FORMAT_OPTIONAL);
 
   switch (op)
   {
+    case 'C':
+      snprintf (tmp, sizeof (tmp), "%%%sd", fmt);
+      snprintf (dest, destlen, tmp, folder->num + 1);
+      break;
+      
     case 'd':
-      if (folder->f != NULL)
+      if (folder->ff->st != NULL)
       {
 	tnow = time (NULL);
-	t_fmt = tnow - folder->f->st_mtime < 31536000 ? "%b %d %H:%M" : "%b %d  %Y";
-	strftime (date, sizeof (date), t_fmt, localtime (&folder->f->st_mtime));
-	snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
-	snprintf (dest, destlen, tmp, date);
+	t_fmt = tnow - folder->ff->st->st_mtime < 31536000 ? "%b %d %H:%M" : "%b %d  %Y";
+	strftime (date, sizeof (date), t_fmt, localtime (&folder->ff->st->st_mtime));
+	mutt_format_s (dest, destlen, fmt, date);
       }
       else
-      {
-	snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
-	snprintf (dest, destlen, tmp, "");
-      }
+	mutt_format_s (dest, destlen, fmt, "");
       break;
+      
     case 'f':
-      strfcpy (fn, folder->name, sizeof(fn));
-      if (folder->f != NULL)
+#ifdef USE_IMAP
+      if (folder->ff->imap)
+        strfcpy (fn, NONULL(folder->ff->desc), sizeof (fn));
+      else
+#endif
+      strfcpy (fn, folder->ff->name, sizeof(fn));
+      if (folder->ff->st != NULL)
       {
-	strcat (fn, S_ISLNK (folder->f->st_mode) ? "@" : 
-		(S_ISDIR (folder->f->st_mode) ? "/" : 
-		 ((folder->f->st_mode & S_IXUSR) != 0 ? "*" : "")));
+	strcat (fn, S_ISLNK (folder->ff->st->st_mode) ? "@" :		/* __STRCAT_CHECKED__ */
+		(S_ISDIR (folder->ff->st->st_mode) ? "/" : 
+		 ((folder->ff->st->st_mode & S_IXUSR) != 0 ? "*" : "")));
       }
-      snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
-      snprintf (dest, destlen, tmp, fn);
+      mutt_format_s (dest, destlen, fmt, fn);
       break;
+      
     case 'F':
-      if (folder->f != NULL)
+      if (folder->ff->st != NULL)
       {
 	sprintf (permission, "%c%c%c%c%c%c%c%c%c%c",
-		 S_ISDIR(folder->f->st_mode) ? 'd' : (S_ISLNK(folder->f->st_mode) ? 'l' : '-'),
-		 (folder->f->st_mode & S_IRUSR) != 0 ? 'r': '-',
-		 (folder->f->st_mode & S_IWUSR) != 0 ? 'w' : '-',
-		 (folder->f->st_mode & S_ISUID) != 0 ? 's' : (folder->f->st_mode & S_IXUSR) != 0 ? 'x': '-',
-		 (folder->f->st_mode & S_IRGRP) != 0 ? 'r' : '-',
-		 (folder->f->st_mode & S_IWGRP) != 0 ? 'w' : '-',
-		 (folder->f->st_mode & S_ISGID) != 0 ? 's' : (folder->f->st_mode & S_IXGRP) != 0 ? 'x': '-',
-		 (folder->f->st_mode & S_IROTH) != 0 ? 'r' : '-',
-		 (folder->f->st_mode & S_IWOTH) != 0 ? 'w' : '-',
-		 (folder->f->st_mode & S_ISVTX) != 0 ? 't' : (folder->f->st_mode & S_IXOTH) != 0 ? 'x': '-');
-	snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
-	snprintf (dest, destlen, tmp, permission);
+	    S_ISDIR(folder->ff->st->st_mode) ? 'd' : (S_ISLNK(folder->ff->st->st_mode) ? 'l' : '-'),
+	    (folder->ff->st->st_mode & S_IRUSR) != 0 ? 'r': '-',
+	    (folder->ff->st->st_mode & S_IWUSR) != 0 ? 'w' : '-',
+	    (folder->ff->st->st_mode & S_ISUID) != 0 ? 's' : (folder->ff->st->st_mode & S_IXUSR) != 0 ? 'x': '-',
+	    (folder->ff->st->st_mode & S_IRGRP) != 0 ? 'r' : '-',
+	    (folder->ff->st->st_mode & S_IWGRP) != 0 ? 'w' : '-',
+	    (folder->ff->st->st_mode & S_ISGID) != 0 ? 's' : (folder->ff->st->st_mode & S_IXGRP) != 0 ? 'x': '-',
+	    (folder->ff->st->st_mode & S_IROTH) != 0 ? 'r' : '-',
+	    (folder->ff->st->st_mode & S_IWOTH) != 0 ? 'w' : '-',
+	    (folder->ff->st->st_mode & S_ISVTX) != 0 ? 't' : (folder->ff->st->st_mode & S_IXOTH) != 0 ? 'x': '-');
+	mutt_format_s (dest, destlen, fmt, permission);
       }
-      else
-      {
 #ifdef USE_IMAP
-	if (strchr(folder->name, '{'))
-	{
-	  snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
-	  snprintf (dest, destlen, tmp, "IMAP");
-	}                                        
-#endif
-      }
-      break;
-    case 'g':
-      if (folder->f != NULL)
+      else if (folder->ff->imap)
       {
-	if ((gr = getgrgid (folder->f->st_gid)))
-	{
-	  snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
-	  snprintf (dest, destlen, tmp, gr->gr_name);
-	}
+	/* mark folders with subfolders AND mail */
+	sprintf (permission, "IMAP %c",
+          (folder->ff->inferiors && folder->ff->selectable) ? '+' : ' ');
+	mutt_format_s (dest, destlen, fmt, permission);
+      }                                        
+#endif
+      else
+	mutt_format_s (dest, destlen, fmt, "");
+      break;
+      
+    case 'g':
+      if (folder->ff->st != NULL)
+      {
+	if ((gr = getgrgid (folder->ff->st->st_gid)))
+	  mutt_format_s (dest, destlen, fmt, gr->gr_name);
 	else
 	{
 	  snprintf (tmp, sizeof (tmp), "%%%sld", fmt);
-	  snprintf (dest, destlen, tmp, folder->f->st_gid);
+	  snprintf (dest, destlen, tmp, folder->ff->st->st_gid);
 	}
       }
       else
-      {
-	snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
-	snprintf (dest, destlen, tmp, "");
-      }
+	mutt_format_s (dest, destlen, fmt, "");
       break;
+      
     case 'l':
-      if (folder->f != NULL)
+      if (folder->ff->st != NULL)
       {
 	snprintf (tmp, sizeof (tmp), "%%%sd", fmt);
-	snprintf (dest, destlen, tmp, folder->f->st_nlink);
+	snprintf (dest, destlen, tmp, folder->ff->st->st_nlink);
       }
       else
-      {
-	snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
-	snprintf (dest, destlen, tmp, "");
-      }
+	mutt_format_s (dest, destlen, fmt, "");
       break;
+      
     case 'N':
+#ifdef USE_IMAP
+      if (mx_is_imap (folder->ff->desc))
+      {
+	if (!optional)
+	{
+	  snprintf (tmp, sizeof (tmp), "%%%sd", fmt);
+	  snprintf (dest, destlen, tmp, folder->ff->new);
+	}
+	else if (!folder->ff->new)
+	  optional = 0;
+	break;
+      }
+#endif
       snprintf (tmp, sizeof (tmp), "%%%sc", fmt);
-      snprintf (dest, destlen, tmp, folder->new ? 'N' : ' ');
+      snprintf (dest, destlen, tmp, folder->ff->new ? 'N' : ' ');
       break;
+      
     case 's':
-      if (folder->f != NULL)
+      if (folder->ff->st != NULL)
       {
 	snprintf (tmp, sizeof (tmp), "%%%sld", fmt);
-	snprintf (dest, destlen, tmp, (long) folder->f->st_size);
+	snprintf (dest, destlen, tmp, (long) folder->ff->st->st_size);
       }
       else
-      {
-	snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
-	snprintf (dest, destlen, tmp, "");
-      }
+	mutt_format_s (dest, destlen, fmt, "");
       break;
+      
+    case 't':
+      snprintf (tmp, sizeof (tmp), "%%%sc", fmt);
+      snprintf (dest, destlen, tmp, folder->ff->tagged ? '*' : ' ');
+      break;
+      
     case 'u':
-      if (folder->f != NULL)
+      if (folder->ff->st != NULL)
       {
-	if ((pw = getpwuid (folder->f->st_uid)))
-	{
-	  snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
-	  snprintf (dest, destlen, tmp, pw->pw_name);
-	}
+	if ((pw = getpwuid (folder->ff->st->st_uid)))
+	  mutt_format_s (dest, destlen, fmt, pw->pw_name);
 	else
 	{
 	  snprintf (tmp, sizeof (tmp), "%%%sld", fmt);
-	  snprintf (dest, destlen, tmp, folder->f->st_uid);
+	  snprintf (dest, destlen, tmp, folder->ff->st->st_uid);
 	}
       }
       else
-      {
-	snprintf (tmp, sizeof (tmp), "%%%ss", fmt);
-	snprintf (dest, destlen, tmp, "");
-      }
+	mutt_format_s (dest, destlen, fmt, "");
+      break;
+      
+    default:
+      snprintf (tmp, sizeof (tmp), "%%%sc", fmt);
+      snprintf (dest, destlen, tmp, op);
       break;
   }
+
+  if (optional)
+    mutt_FormatString (dest, destlen, ifstring, folder_format_str, data, 0);
+  else if (flags & M_FORMAT_OPTIONAL)
+    mutt_FormatString (dest, destlen, elsestring, folder_format_str, data, 0);
+
   return (src);
 }
 
 static void add_folder (MUTTMENU *m, struct browser_state *state,
 			const char *name, const struct stat *s, int new)
 {
-  char buffer[_POSIX_PATH_MAX + SHORT_STRING];
-  FOLDER folder;
-
-  folder.name = name;
-  folder.f = s;
-  folder.new = new;
-  mutt_FormatString (buffer, sizeof (buffer), NONULL(FolderFormat),
-		     folder_format_str, (unsigned long) &folder,
-		     M_FORMAT_ARROWCURSOR);
-
   if (state->entrylen == state->entrymax)
   {
     /* need to allocate more space */
     safe_realloc ((void **) &state->entry,
 		  sizeof (struct folder_file) * (state->entrymax += 256));
+    memset (&state->entry[state->entrylen], 0,
+	    sizeof (struct folder_file) * 256);
     if (m)
       m->data = state->entry;
   }
 
   if (s != NULL)
   {
-  (state->entry)[state->entrylen].mode = s->st_mode;
-  (state->entry)[state->entrylen].mtime = s->st_mtime;
-  (state->entry)[state->entrylen].size = s->st_size;
+    (state->entry)[state->entrylen].mode = s->st_mode;
+    (state->entry)[state->entrylen].mtime = s->st_mtime;
+    (state->entry)[state->entrylen].size = s->st_size;
+    
+    (state->entry)[state->entrylen].st = safe_malloc (sizeof (struct stat));
+    memcpy ((state->entry)[state->entrylen].st, s, sizeof (struct stat));
   }
-  (state->entry)[state->entrylen].name = safe_strdup (name);
-  (state->entry)[state->entrylen].desc = safe_strdup (buffer);
 
+  (state->entry)[state->entrylen].new = new;
+  (state->entry)[state->entrylen].name = safe_strdup (name);
+  (state->entry)[state->entrylen].desc = safe_strdup (name);
+#ifdef USE_IMAP
+  (state->entry)[state->entrylen].imap = 0;
+#endif
   (state->entrylen)++;
 }
 
@@ -319,13 +336,16 @@ static void init_state (struct browser_state *state, MUTTMENU *menu)
 {
   state->entrylen = 0;
   state->entrymax = 256;
-  state->entry = (struct folder_file *) safe_malloc (sizeof (struct folder_file) * state->entrymax);
+  state->entry = (struct folder_file *) safe_calloc (state->entrymax, sizeof (struct folder_file));
+#ifdef USE_IMAP
+  state->imap_browse = 0;
+#endif
   if (menu)
     menu->data = state->entry;
 }
 
 static int examine_directory (MUTTMENU *menu, struct browser_state *state,
-			      const char *d, const char *prefix)
+			      char *d, const char *prefix)
 {
   struct stat s;
   DIR *dp;
@@ -333,8 +353,19 @@ static int examine_directory (MUTTMENU *menu, struct browser_state *state,
   char buffer[_POSIX_PATH_MAX + SHORT_STRING];
   BUFFY *tmp;
 
-  if (stat (d, &s) == -1)
+  while (stat (d, &s) == -1)
   {
+    if (errno == ENOENT)
+    {
+      /* The last used directory is deleted, try to use the parent dir. */
+      char *c = strrchr (d, '/');
+
+      if (c && (c > d))
+      {
+	*c = 0;
+	continue;
+      }
+    }
     mutt_perror (d);
     return (-1);
   }
@@ -398,7 +429,14 @@ static int examine_mailboxes (MUTTMENU *menu, struct browser_state *state)
   do
   {
 #ifdef USE_IMAP
-    if (tmp->path[0] == '{')
+    if (mx_is_imap (tmp->path))
+    {
+      add_folder (menu, state, tmp->path, NULL, tmp->new);
+      continue;
+    }
+#endif
+#ifdef USE_POP
+    if (mx_is_pop (tmp->path))
     {
       add_folder (menu, state, tmp->path, NULL, tmp->new);
       continue;
@@ -428,7 +466,13 @@ int select_file_search (MUTTMENU *menu, regex_t *re, int n)
 
 void folder_entry (char *s, size_t slen, MUTTMENU *menu, int num)
 {
-  snprintf (s, slen, "%2d %s", num + 1, ((struct folder_file *) menu->data)[num].desc);
+  FOLDER folder;
+
+  folder.ff = &((struct folder_file *) menu->data)[num];
+  folder.num = num;
+  
+  mutt_FormatString (s, slen, NONULL(FolderFormat), folder_format_str, 
+      (unsigned long) &folder, M_FORMAT_ARROWCURSOR);
 }
 
 static void init_menu (struct browser_state *state, MUTTMENU *menu, char *title,
@@ -445,19 +489,43 @@ static void init_menu (struct browser_state *state, MUTTMENU *menu, char *title,
   if (menu->top > menu->current)
     menu->top = 0;
 
+  menu->tagged = 0;
+  
   if (buffy)
     snprintf (title, titlelen, _("Mailboxes [%d]"), mutt_buffy_check (0));
   else
   {
     strfcpy (path, LastDir, sizeof (path));
     mutt_pretty_mailbox (path);
+#ifdef USE_IMAP
+  if (state->imap_browse && option (OPTIMAPLSUB))
+    snprintf (title, titlelen, _("Subscribed [%s], File mask: %s"),
+	      path, NONULL (Mask.pattern));
+  else
+#endif
     snprintf (title, titlelen, _("Directory [%s], File mask: %s"),
-	      path, Mask.pattern);
+	      path, NONULL(Mask.pattern));
   }
   menu->redraw = REDRAW_FULL;
 }
 
-void mutt_select_file (char *f, size_t flen, int buffy)
+int file_tag (MUTTMENU *menu, int n, int m)
+{
+  struct folder_file *ff = &(((struct folder_file *)menu->data)[n]);
+  int ot;
+  if (S_ISDIR (ff->mode) || (S_ISLNK (ff->mode) && link_is_dir (LastDir, ff->name)))
+  {
+    mutt_error _("Can't attach a directory!");
+    return 0;
+  }
+  
+  ot = ff->tagged;
+  ff->tagged = (m >= 0 ? m : !ff->tagged);
+  
+  return ff->tagged - ot;
+}
+
+void _mutt_select_file (char *f, size_t flen, int flags, char ***files, int *numfiles)
 {
   char buf[_POSIX_PATH_MAX];
   char prefix[_POSIX_PATH_MAX] = "";
@@ -467,12 +535,31 @@ void mutt_select_file (char *f, size_t flen, int buffy)
   MUTTMENU *menu;
   struct stat st;
   int i, killPrefix = 0;
+  int multiple = (flags & M_SEL_MULTI)  ? 1 : 0;
+  int folder   = (flags & M_SEL_FOLDER) ? 1 : 0;
+  int buffy    = (flags & M_SEL_BUFFY)  ? 1 : 0;
 
+  buffy = buffy && folder;
+  
   memset (&state, 0, sizeof (struct browser_state));
 
+  if (!folder)
+    strfcpy (LastDirBackup, LastDir, sizeof (LastDirBackup));
+
   if (*f)
   {
     mutt_expand_path (f, flen);
+#ifdef USE_IMAP
+    if (mx_is_imap (f))
+    {
+      init_state (&state, NULL);
+      state.imap_browse = 1;
+      imap_browse (f, &state);
+      strfcpy (LastDir, state.folder, sizeof (LastDir));
+    }
+    else
+    {
+#endif
     for (i = mutt_strlen (f) - 1; i > 0 && f[i] != '/' ; i--);
     if (i > 0)
     {
@@ -485,14 +572,14 @@ void mutt_select_file (char *f, size_t flen, int buffy)
       else
       {
 	getcwd (LastDir, sizeof (LastDir));
-	strcat (LastDir, "/");
+	strcat (LastDir, "/");	/* __STRCAT_CHECKED__ */
 	strncat (LastDir, f, i);
       }
     }
     else
     {
       if (f[0] == '/')
-	strcpy (LastDir, "/");
+	strcpy (LastDir, "/");		/* __STRCPY_CHECKED__ */
       else
 	getcwd (LastDir, sizeof (LastDir));
     }
@@ -502,19 +589,40 @@ void mutt_select_file (char *f, size_t flen, int buffy)
     else
       strfcpy (prefix, f + i + 1, sizeof (prefix));
     killPrefix = 1;
+#ifdef USE_IMAP
+    }
+#endif
+  }
+  else 
+  {
+    if (!folder)
+      getcwd (LastDir, sizeof (LastDir));
+    else if (!LastDir[0])
+      strfcpy (LastDir, NONULL(Maildir), sizeof (LastDir));
+    
+#ifdef USE_IMAP
+    if (!buffy && mx_is_imap (LastDir))
+    {
+      init_state (&state, NULL);
+      state.imap_browse = 1;
+      imap_browse (LastDir, &state);
+    }
+#endif
   }
-  else if (!LastDir[0])
-    strfcpy (LastDir, NONULL(Maildir), sizeof (LastDir));
 
   *f = 0;
 
   if (buffy)
   {
     if (examine_mailboxes (NULL, &state) == -1)
-      return;
+      goto bail;
   }
-  else if (examine_directory (NULL, &state, LastDir, prefix) == -1)
-    return;
+  else
+#ifdef USE_IMAP
+  if (!state.imap_browse)
+#endif
+  if (examine_directory (NULL, &state, LastDir, prefix) == -1)
+    goto bail;
 
   menu = mutt_new_menu ();
   menu->menu = MENU_FOLDER;
@@ -522,8 +630,11 @@ void mutt_select_file (char *f, size_t flen, int buffy)
   menu->search = select_file_search;
   menu->title = title;
   menu->data = state.entry;
+  if (multiple)
+    menu->tag = file_tag;
 
-  menu->help = mutt_compile_help (helpstr, sizeof (helpstr), MENU_FOLDER, FolderHelp);
+  menu->help = mutt_compile_help (helpstr, sizeof (helpstr), MENU_FOLDER,
+    FolderHelp);
 
   init_menu (&state, menu, title, sizeof (title), buffy);
 
@@ -541,7 +652,11 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 
         if (S_ISDIR (state.entry[menu->current].mode) ||
 	    (S_ISLNK (state.entry[menu->current].mode) &&
-	    link_is_dir (state.entry[menu->current].name)))
+	    link_is_dir (LastDir, state.entry[menu->current].name)) 
+#ifdef USE_IMAP
+	    || state.entry[menu->current].inferiors
+#endif
+	    )
 	{
 	  /* make sure this isn't a MH or maildir mailbox */
 	  if (buffy)
@@ -549,11 +664,21 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 	    strfcpy (buf, state.entry[menu->current].name, sizeof (buf));
 	    mutt_expand_path (buf, sizeof (buf));
 	  }
+#ifdef USE_IMAP
+	  else if (state.imap_browse)
+	  {
+            strfcpy (buf, state.entry[menu->current].name, sizeof (buf));
+	  }
+#endif
 	  else
 	    snprintf (buf, sizeof (buf), "%s/%s", LastDir, 
 		      state.entry[menu->current].name);
 
-	  if (mx_get_magic (buf) <= 0)
+	  if ((mx_get_magic (buf) <= 0)
+#ifdef USE_IMAP
+	    || state.entry[menu->current].inferiors
+#endif
+	    )
 	  {
 	    char OldLastDir[_POSIX_PATH_MAX];
 
@@ -563,7 +688,7 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 	    if (mutt_strcmp (state.entry[menu->current].name, "..") == 0)
 	    {
 	      if (mutt_strcmp ("..", LastDir + mutt_strlen (LastDir) - 2) == 0)
-		strcat (LastDir, "/..");
+		strcat (LastDir, "/..");	/* __STRCAT_CHECKED__ */
 	      else
 	      {
 		char *p = strrchr (LastDir + 1, '/');
@@ -575,18 +700,43 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 		  if (LastDir[0] == '/')
 		    LastDir[1] = 0;
 		  else
-		    strcat (LastDir, "/..");
+		    strcat (LastDir, "/..");	/* __STRCAT_CHECKED__ */
 		}
 	      }
 	    }
 	    else if (buffy)
 	    {
-	      sprintf (LastDir, "%s", state.entry[menu->current].name);
+	      strfcpy (LastDir, state.entry[menu->current].name, sizeof (LastDir));
 	      mutt_expand_path (LastDir, sizeof (LastDir));
 	    }
+#ifdef USE_IMAP
+	    else if (state.imap_browse)
+	    {
+	      int n;
+	      ciss_url_t url;
+	      
+              strfcpy (LastDir, state.entry[menu->current].name,
+                sizeof (LastDir));
+	      /* tack on delimiter here */
+	      n = strlen (LastDir)+1;
+	      
+	      /* special case "" needs no delimiter */
+	      url_parse_ciss (&url, state.entry[menu->current].name);
+	      if (url.path &&
+		  (state.entry[menu->current].delim != '\0') &&
+		  (n < sizeof (LastDir)))
+	      {
+		LastDir[n] = '\0';
+		LastDir[n-1] = state.entry[menu->current].delim;
+	      }
+	    }
+#endif
 	    else
-	      sprintf (LastDir + mutt_strlen (LastDir), "/%s",
-		       state.entry[menu->current].name);
+	    {
+	      char tmp[_POSIX_PATH_MAX];
+	      snprintf (tmp, sizeof (tmp), "%s/%s", LastDir, state.entry[menu->current].name);
+	      strfcpy (LastDir, tmp, sizeof (LastDir));
+	    }
 
 	    destroy_state (&state);
 	    if (killPrefix)
@@ -595,6 +745,16 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 	      killPrefix = 0;
 	    }
 	    buffy = 0;
+#ifdef USE_IMAP
+	    if (state.imap_browse)
+	    {
+	      init_state (&state, NULL);
+	      state.imap_browse = 1;
+	      imap_browse (LastDir, &state);
+	      menu->data = state.entry;
+	    }
+	    else
+#endif
 	    if (examine_directory (menu, &state, LastDir, prefix) == -1)
 	    {
 	      /* try to restore the old values */
@@ -602,7 +762,7 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 	      if (examine_directory (menu, &state, LastDir, prefix) == -1)
 	      {
 		strfcpy (LastDir, NONULL(Homedir), sizeof (LastDir));
-		return;
+		goto bail;
 	      }
 	    }
 	    menu->current = 0; 
@@ -617,6 +777,10 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 	  strfcpy (f, state.entry[menu->current].name, flen);
 	  mutt_expand_path (f, flen);
 	}
+#ifdef USE_IMAP
+	else if (state.imap_browse)
+          strfcpy (f, state.entry[menu->current].name, flen);
+#endif
 	else
 	  snprintf (f, flen, "%s/%s", LastDir, state.entry[menu->current].name);
 
@@ -624,22 +788,135 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 
       case OP_EXIT:
 
+	if (multiple)
+	{
+	  char **tfiles;
+	  int i, j;
+
+	  if (menu->tagged)
+	  {
+	    *numfiles = menu->tagged;
+	    tfiles = safe_malloc (*numfiles * sizeof (char *));
+	    for (i = 0, j = 0; i < state.entrylen; i++)
+	    {
+	      struct folder_file ff = state.entry[i];
+	      char full[_POSIX_PATH_MAX];
+	      if (ff.tagged)
+	      {
+		snprintf (full, sizeof (full), "%s/%s", LastDir, ff.name);
+		mutt_expand_path (full, sizeof (full));
+		tfiles[j++] = safe_strdup (full);
+	      }
+	    }
+	    *files = tfiles;
+	  }
+	  else if (f[0]) /* no tagged entries. return selected entry */
+	  {
+	    *numfiles = 1;
+	    tfiles = safe_malloc (*numfiles * sizeof (char *));
+	    mutt_expand_path (f, flen);
+	    tfiles[0] = safe_strdup (f);
+	    *files = tfiles;
+	  }
+	}
+
 	destroy_state (&state);
 	mutt_menuDestroy (&menu);
-	return;
+	goto bail;
 
       case OP_BROWSER_TELL:
         if(state.entrylen)
-	  mutt_message(state.entry[menu->current].name);
+	  mutt_message("%s", state.entry[menu->current].name);
         break;
+
+#ifdef USE_IMAP
+      case OP_BROWSER_SUBSCRIBE:
+	imap_subscribe (state.entry[menu->current].name, 1);
+	break;
+
+      case OP_BROWSER_UNSUBSCRIBE:
+	imap_subscribe (state.entry[menu->current].name, 0);
+	break;
+
+      case OP_BROWSER_TOGGLE_LSUB:
+	if (option (OPTIMAPLSUB))
+	{
+	  unset_option (OPTIMAPLSUB);
+	}
+	else
+	{
+	  set_option (OPTIMAPLSUB);
+	}
+	mutt_ungetch (0, OP_CHECK_NEW);
+	break;
+
+      case OP_CREATE_MAILBOX:
+	if (!state.imap_browse)
+	  mutt_error (_("Create is only supported for IMAP mailboxes"));
+	else
+	{
+	  imap_mailbox_create (LastDir);
+	  /* TODO: find a way to detect if the new folder would appear in
+	   *   this window, and insert it without starting over. */
+	  destroy_state (&state);
+	  init_state (&state, NULL);
+	  state.imap_browse = 1;
+	  imap_browse (LastDir, &state);
+	  menu->data = state.entry;
+	  menu->current = 0; 
+	  menu->top = 0; 
+	  init_menu (&state, menu, title, sizeof (title), buffy);
+	  MAYBE_REDRAW (menu->redraw);
+	}
+	break;
+
+      case OP_DELETE_MAILBOX:
+	if (!state.entry[menu->current].imap)
+	  mutt_error (_("Delete is only supported for IMAP mailboxes"));
+	else
+        {
+	  char msg[SHORT_STRING];
+	  IMAP_MBOX mx;
+	  int nentry = menu->current;
+	  
+	  imap_parse_path (state.entry[nentry].name, &mx);
+	  snprintf (msg, sizeof (msg), _("Really delete mailbox \"%s\"?"),
+            mx.mbox);
+	  if (mutt_yesorno (msg, M_NO) == M_YES)
+          {
+	    if (!imap_delete_mailbox (Context, mx))
+            {
+	      /* free the mailbox from the browser */
+	      safe_free ((void **) &((state.entry)[nentry].name));
+	      safe_free ((void **) &((state.entry)[nentry].desc));
+	      /* and move all other entries up */
+	      if (nentry+1 < state.entrylen)
+		memmove (state.entry + nentry, state.entry + nentry + 1,
+                  sizeof (struct folder_file) * (state.entrylen - (nentry+1)));
+	      state.entrylen--;
+	      mutt_message _("Mailbox deleted.");
+	      init_menu (&state, menu, title, sizeof (title), buffy);
+	      MAYBE_REDRAW (menu->redraw);
+	    }
+	  }
+	  else
+	    mutt_message _("Mailbox not deleted.");
+	  FREE (&mx.mbox);
+        }
+        break;
+#endif
       
       case OP_CHANGE_DIRECTORY:
 
 	strfcpy (buf, LastDir, sizeof (buf));
-	{/* add '/' at the end of the directory name */
-	int len=mutt_strlen(LastDir);
-	if (sizeof (buf) > len)
-	  buf[len]='/';
+#ifdef USE_IMAP
+	if (!state.imap_browse)
+#endif
+	{
+	  /* add '/' at the end of the directory name if not already there */
+	  int len=mutt_strlen(LastDir);
+	  if (len && LastDir[len-1] != '/' && sizeof (buf) > len)
+	    buf[len]='/';
 	}
 
 	if (mutt_get_field (_("Chdir to: "), buf, sizeof (buf), M_FILE) == 0 &&
@@ -647,6 +924,21 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 	{
 	  buffy = 0;	  
 	  mutt_expand_path (buf, sizeof (buf));
+#ifdef USE_IMAP
+	  if (mx_is_imap (buf))
+	  {
+	    strfcpy (LastDir, buf, sizeof (LastDir));
+	    destroy_state (&state);
+	    init_state (&state, NULL);
+	    state.imap_browse = 1;
+	    imap_browse (LastDir, &state);
+	    menu->data = state.entry;
+	    menu->current = 0; 
+	    menu->top = 0; 
+	    init_menu (&state, menu, title, sizeof (title), buffy);
+	  }
+	  else
+#endif
 	  if (stat (buf, &st) == 0)
 	  {
 	    if (S_ISDIR (st.st_mode))
@@ -664,7 +956,7 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 		mutt_error _("Error scanning directory.");
 		destroy_state (&state);
 		mutt_menuDestroy (&menu);
-		return;
+		goto bail;
 	      }
 	    }
 	    else
@@ -706,21 +998,31 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 	  }
 	  else
 	  {
-	    safe_free ((void **) &Mask.pattern);
+	    mutt_str_replace (&Mask.pattern, buf);
 	    regfree (Mask.rx);
 	    safe_free ((void **) &Mask.rx);
-	    Mask.pattern = safe_strdup (buf);
 	    Mask.rx = rx;
 	    Mask.not = not;
 
 	    destroy_state (&state);
+#ifdef USE_IMAP
+	    if (state.imap_browse)
+	    {
+	      init_state (&state, NULL);
+	      state.imap_browse = 1;
+	      imap_browse (LastDir, &state);
+	      menu->data = state.entry;
+	      init_menu (&state, menu, title, sizeof (title), buffy);
+	    }
+	    else
+#endif
 	    if (examine_directory (menu, &state, LastDir, NULL) == 0)
 	      init_menu (&state, menu, title, sizeof (title), buffy);
 	    else
 	    {
 	      mutt_error _("Error scanning directory.");
 	      mutt_menuDestroy (&menu);
-	      return;
+	      goto bail;
 	    }
 	    killPrefix = 0;
 	    if (!state.entrylen)
@@ -782,13 +1084,23 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 	destroy_state (&state);
 	prefix[0] = 0;
 	killPrefix = 0;
+
 	if (buffy)
 	{
 	  if (examine_mailboxes (menu, &state) == -1)
-	    return;
+	    goto bail;
 	}
+#ifdef USE_IMAP
+	else if (mx_is_imap (LastDir))
+	{
+	  init_state (&state, NULL);
+	  state.imap_browse = 1;
+	  imap_browse (LastDir, &state);
+	  menu->data = state.entry;
+	}
+#endif
 	else if (examine_directory (menu, &state, LastDir, prefix) == -1)
-	  return;
+	  goto bail;
 	init_menu (&state, menu, title, sizeof (title), buffy);
 	break;
 
@@ -800,7 +1112,7 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 	  strfcpy (f, buf, flen);
 	  destroy_state (&state);
 	  mutt_menuDestroy (&menu);
-	  return;
+	  goto bail;
 	}
 	MAYBE_REDRAW (menu->redraw);
 	break;
@@ -812,9 +1124,19 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 	  break;
 	}
 
+#ifdef USE_IMAP
+	if (state.entry[menu->current].selectable)
+	{
+	  strfcpy (f, state.entry[menu->current].name, flen);
+	  destroy_state (&state);
+	  mutt_menuDestroy (&menu);
+	  goto bail;
+	}
+	else
+#endif
         if (S_ISDIR (state.entry[menu->current].mode) ||
 	    (S_ISLNK (state.entry[menu->current].mode) &&
-	    link_is_dir (state.entry[menu->current].name)))
+	    link_is_dir (LastDir, state.entry[menu->current].name)))
 	{
 	  mutt_error _("Can't view a directory");
 	  break;
@@ -829,7 +1151,7 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 	  b = mutt_make_file_attach (buf);
 	  if (b != NULL)
 	  {
-	    mutt_view_attachment (NULL, b, M_REGULAR);
+	    mutt_view_attachment (NULL, b, M_REGULAR, NULL, NULL, 0);
 	    mutt_free_body (&b);
 	    menu->redraw = REDRAW_FULL;
 	  }
@@ -838,5 +1160,10 @@ void mutt_select_file (char *f, size_t flen, int buffy)
 	}
     }
   }
-  /* not reached */
+  
+  bail:
+  
+  if (!folder)
+    strfcpy (LastDir, LastDirBackup, sizeof (LastDir));
+  
 }
