@@ -1,4 +1,4 @@
-/* $Id: browser.h,v 2.7 2000/03/03 10:10:06 roessler Exp $ */
+/* $Id: browser.h,v 2.9 2000/08/21 15:42:20 roessler Exp $ */
 /*
  * Copyright (C) 1996-2000 Michael R. Elkins <me@cs.hmc.edu>
  * 
@@ -29,6 +29,8 @@ struct folder_file
 
   char *name;
   char *desc;
+
+  unsigned short new;
 #ifdef USE_IMAP
   char delim;
   
@@ -37,7 +39,6 @@ struct folder_file
   unsigned inferiors : 1;
 #endif
   unsigned tagged : 1;
-  unsigned is_new : 1;
 };
 
 struct browser_state
@@ -48,9 +49,9 @@ struct browser_state
 #ifdef USE_IMAP
   short imap_browse;
   char *folder;
-  int noselect : 1;
-  int marked : 1;
-  int unmarked : 1;
+  unsigned noselect : 1;
+  unsigned marked : 1;
+  unsigned unmarked : 1;
 #endif
 };
 
