@@ -93,7 +93,7 @@ static int edit_one_message (CONTEXT *ctx, HEADER *cur)
 
   if ((rc = stat (tmp, &sb)) == -1)
   {
-    mutt_error (_("Can't stat: %s"), strerror (errno));
+    mutt_error (_("Can't stat %s: %s"), tmp, strerror (errno));
     goto bail;
   }
   
@@ -127,7 +127,7 @@ static int edit_one_message (CONTEXT *ctx, HEADER *cur)
 
   of = cf = 0;
   
-  if (fgets (buff, sizeof (buff), fp) && is_from (buff, NULL, 0))
+  if (fgets (buff, sizeof (buff), fp) && is_from (buff, NULL, 0, NULL))
   {
     if (tmpctx.magic == M_MBOX || tmpctx.magic == M_MMDF ||
 	tmpctx.magic == M_KENDRA)
