@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1996-8 Michael R. Elkins <me@cs.hmc.edu>
+ * Copyright (C) 1996-2000 Michael R. Elkins <me@cs.hmc.edu>
  * 
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -13,7 +13,7 @@
  * 
  *     You should have received a copy of the GNU General Public License
  *     along with this program; if not, write to the Free Software
- *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
  */ 
 
 #include <stdlib.h>
@@ -33,11 +33,14 @@
 
 
 
-#ifdef _PGPPATH
+#ifdef HAVE_PGP
 #include "pgp.h"
 #endif
 
 
+#define BUFI_SIZE 1000
+#define BUFO_SIZE 2000
+
 
 typedef void handler_f (BODY *, STATE *);
 typedef handler_f *handler_t;
@@ -64,33 +67,64 @@ int Index_64[128] = {
     41,42,43,44, 45,46,47,48, 49,50,51,-1, -1,-1,-1,-1
 };
 
-static void state_maybe_utf8_putc(STATE *s, char c, int is_utf8, CHARSET *chs, CHARSET_MAP *map)
+static void state_prefix_put (const char *d, size_t dlen, STATE *s)
 {
-  if(is_utf8)
-    state_fput_utf8(s, c, chs);
+  if (s->prefix)
+    while (dlen--)
+      state_prefix_putc (*d++, s);
   else
-    state_prefix_putc(mutt_display_char ((unsigned char) c, map), s);
+    fwrite (d, dlen, 1, s->fpout);
 }
 
-void mutt_decode_xbit (STATE *s, BODY *b, int istext)
+static void convert_to_state(iconv_t cd, char *bufi, size_t *l, STATE *s)
+{
+  char bufo[BUFO_SIZE];
+  const char *ib;
+  char *ob;
+  size_t ibl, obl;
+
+  if (!bufi)
+  {
+    if (cd != (iconv_t)(-1))
+    {
+      ob = bufo, obl = sizeof (bufo);
+      iconv (cd, 0, 0, &ob, &obl);
+      if (ob != bufo)
+	state_prefix_put (bufo, ob - bufo, s);
+    }
+    return;
+  }
+
+  if (cd == (iconv_t)(-1))
+  {
+    state_prefix_put (bufi, *l, s);
+    *l = 0;
+    return;
+  }
+
+  ib = bufi, ibl = *l;
+  for (;;)
+  {
+    ob = bufo, obl = sizeof (bufo);
+    mutt_iconv (cd, &ib, &ibl, &ob, &obl, 0, "?");
+    if (ob == bufo)
+      break;
+    state_prefix_put (bufo, ob - bufo, s);
+  }
+  memmove (bufi, ib, ibl);
+  *l = ibl;
+}
+
+void mutt_decode_xbit (STATE *s, long len, int istext, iconv_t cd)
 {
-  long len = b->length;
   int c, ch;
-  
+  char bufi[BUFI_SIZE];
+  size_t l = 0;
+
   if (istext)
   {
-    CHARSET_MAP *map = NULL;
-    CHARSET *chs = NULL;
-    char *charset = mutt_get_parameter("charset", b->parameter);
-    int is_utf8;
-
-    if((is_utf8 = (mutt_is_utf8(charset) && !mutt_is_utf8(Charset))))
-      chs = mutt_get_charset(Charset);
-    else
-      map = mutt_get_translation(charset, Charset);
-
     state_set_prefix(s);
-    
+
     while ((c = fgetc(s->fpin)) != EOF && len--)
     {
       if(c == '\r' && len)
@@ -103,134 +137,156 @@ void mutt_decode_xbit (STATE *s, BODY *b, int istext)
 	else 
 	  ungetc(ch, s->fpin);
       }
-	
-      state_maybe_utf8_putc(s, c, is_utf8, chs, map);
-    }
-    
-    state_reset_prefix(s);
 
-    if(is_utf8)
-      state_fput_utf8(s, '\0', chs);
-    
+      bufi[l++] = c;
+      if (l == sizeof (bufi))
+	convert_to_state (cd, bufi, &l, s);
+    }
+
+    convert_to_state (cd, bufi, &l, s);
+    convert_to_state (cd, 0, 0, s);
+
+    state_reset_prefix (s);
   }
   else
     mutt_copy_bytes (s->fpin, s->fpout, len);
 }
 
-static int handler_state_fgetc(STATE *s)
+static int qp_decode_triple (char *s, char *d)
 {
-  int ch;
+  /* soft line break */
+  if (*s == '=' && !(*(s+1)))
+    return 1;
   
-  if((ch = fgetc(s->fpin)) == EOF)
+  /* quoted-printable triple */
+  if (*s == '=' && isxdigit (*(s+1)) && isxdigit (*(s+2)))
   {
-    dprint(1, (debugfile, "handler_state_fgetc: unexpected EOF.\n"));
-    state_puts (_("[-- Error: unexpected end of file! --]\n"), s);
+    *d = (hexval (*(s+1)) << 4) | hexval (*(s+2));
+    return 0;
   }
-  return ch;
+  
+  /* something else */
+  return -1;
 }
 
-void mutt_decode_quoted (STATE *s, BODY *b, int istext)
+static void qp_decode_line (char *dest, char *src, size_t *l,
+			    int last)
 {
-  long len = b->length;
-  int ch;
-  char *charset = mutt_get_parameter("charset", b->parameter);
-  int is_utf8 = 0;
-  CHARSET *chs = NULL;
-  CHARSET_MAP *map = NULL;
-  
-  if(istext)
+  char *d, *s;
+  char c;
+
+  int kind;
+  int soft = 0;
+
+  /* chop trailing whitespace */
+
+  if (*src)
   {
-    if((is_utf8 = (mutt_is_utf8(charset) && !mutt_is_utf8(Charset))))
-      chs = mutt_get_charset(Charset);
-    else
-      map = mutt_get_translation(charset, Charset);
+    s = src + strlen(src) - 1;
+    while (s >= src && ISSPACE(*s))
+      s--;
+    *(++s) = '\0';
   }
+
+  /* decode the line */
   
-  state_set_prefix(s);
+  for (d = dest, s = src; *s;)
+  {
+    switch ((kind = qp_decode_triple (s, &c)))
+    {
+      case  0: *d++ = c; s += 3; break;	/* qp triple */
+      case -1: *d++ = *s++;      break; /* single character */
+      case  1: soft = 1; s++;	 break; /* soft line break */
+    }
+  }
+
+  if (!soft && last == '\n')
+    *d++ = '\n';
   
+  *d = '\0';
+  *l = d - dest;
+}
+
+/* 
+ * Decode an attachment encoded with quoted-printable.
+ * 
+ * Why doesn't this overflow any buffers?  First, it's guaranteed
+ * that the length of a line grows when you _en_-code it to
+ * quoted-printable.  That means that we always can store the
+ * result in a buffer of at most the _same_ size.
+ * 
+ * Now, we don't special-case if the line we read with fgets()
+ * isn't terminated.  We don't care about this, since STRING > 78,
+ * so corrupted input will just be corrupted a bit more.  That
+ * implies that STRING+1 bytes are always sufficient to store the
+ * result of qp_decode_line.
+ * 
+ * Finally, at soft line breaks, some part of a multibyte character
+ * may have been left over by convert_to_state().  This shouldn't
+ * be more than 6 characters, so STRING + 7 should be sufficient
+ * memory to store the decoded data.
+ * 
+ * Just to make sure that I didn't make some off-by-one error
+ * above, we just use STRING*2 for the target buffer's size.
+ * 
+ */
+
+void mutt_decode_quoted (STATE *s, long len, int istext, iconv_t cd)
+{
+  char line[STRING];
+  char decline[2*STRING];
+  size_t l = 0;
+  size_t l2;
+  size_t l3;
+  
+  int last;
+  
+  if (istext)
+    state_set_prefix(s);
+
   while (len > 0)
   {
-    if ((ch = handler_state_fgetc(s)) == EOF)
+    last = 0;
+    
+    if (fgets (line, MIN (sizeof (line), len + 1), s->fpin) == NULL)
       break;
 
-    len--;
-    
-    if (ch == '=')
-    {
-      int ch1, ch2;
-      
-      if(!len || (ch1 = handler_state_fgetc(s)) == EOF)
-	break;
+    len -= (l2 = strlen (line));
 
-      len--;
-      
-      if (ch1 == '\n' || ch1 == '\r' || ch1 == ' ' || ch1 == '\t')
-      {
-	/* Skip whitespace at the end of the line since MIME does not
-	 * allow for it
-	 */
-	if(ch1 != '\n')
-	{
-	  while(len && (ch1 = handler_state_fgetc(s)) != EOF)
-	  {
-	    len--;
-	    if(ch1 == '\n')
-	      break;
-	  }
-	}
-
-	if(ch1 == EOF)
-	  break;
-
-	ch = EOF;
-
-      }
-      else
-      {
-	if(!len || (ch2 = handler_state_fgetc(s)) == EOF)
-	  break;
+    /* 
+     * Skip over pending input data until end of line - at this
+     * point, the only thing we may see is trailing whitespace,
+     * i.e. garbage.
+     */
 
+    if (l2 && (last = line[l2 - 1]) != '\n')
+      while (len > 0 && (last = fgetc (s->fpin)) != '\n')
 	len--;
-	
-        ch = hexval (ch1) << 4;
-        ch |= hexval (ch2);
-      }
-    } /* ch == '=' */
-    else if (istext && ch == '\r')
-    {
-      continue;
-    }
-    if(ch != EOF)
-      state_maybe_utf8_putc(s, ch, is_utf8, chs, map);
-  }
-  
-  state_reset_prefix(s);
 
-  if(is_utf8)
-    state_fput_utf8(s, '\0', chs);
+    /* 
+     * decode and do character set conversion
+     */
+    
+    qp_decode_line (decline + l, line, &l3, last);
+    l += l3;
+    convert_to_state (cd, decline, &l, s);
+  }
+
+  convert_to_state (cd, 0, 0, s);
+  state_reset_prefix(s);
 }
 
-void mutt_decode_base64 (STATE *s, BODY *b, int istext)
+void mutt_decode_base64 (STATE *s, long len, int istext, iconv_t cd)
 {
-  long len = b->length;
   char buf[5];
   int c1, c2, c3, c4, ch, cr = 0, i;
-  char *charset = mutt_get_parameter("charset", b->parameter);
-  CHARSET_MAP *map = NULL;
-  CHARSET *chs = NULL;
-  int is_utf8 = 0;
+  char bufi[BUFI_SIZE];
+  size_t l = 0;
 
-  if(istext)
-  {
-    if((is_utf8 = (mutt_is_utf8(charset) && !mutt_is_utf8(Charset))))
-      chs = mutt_get_charset(Charset);
-    else
-      map = mutt_get_translation(charset, Charset);
-  }
-  
   buf[4] = 0;
 
-  if (istext) state_set_prefix(s);
+  if (istext) 
+    state_set_prefix(s);
 
   while (len > 0)
   {
@@ -242,20 +298,25 @@ void mutt_decode_base64 (STATE *s, BODY *b, int istext)
 	buf[i++] = ch;
     }
     if (i != 4)
-      return; /* didn't get a multiple of four chars! */
+    {
+      dprint (2, (debugfile, "%s:%d [mutt_decode_base64()]: "
+		  "didn't get a multiple of 4 chars.\n", __FILE__, __LINE__));
+      break;
+    }
 
     c1 = base64val (buf[0]);
     c2 = base64val (buf[1]);
     ch = (c1 << 2) | (c2 >> 4);
 
     if (cr && ch != '\n') 
-      state_maybe_utf8_putc(s, '\r', is_utf8, chs, map);
+      bufi[l++] = '\r';
+
     cr = 0;
       
     if (istext && ch == '\r')
       cr = 1;
     else
-      state_maybe_utf8_putc(s, ch, is_utf8, chs, map);
+      bufi[l++] = ch;
 
     if (buf[2] == '=')
       break;
@@ -263,28 +324,37 @@ void mutt_decode_base64 (STATE *s, BODY *b, int istext)
     ch = ((c2 & 0xf) << 4) | (c3 >> 2);
 
     if (cr && ch != '\n')
-      state_maybe_utf8_putc(s, ch, is_utf8, chs, map);
+      bufi[l++] = '\r';
 
     cr = 0;
 
     if (istext && ch == '\r')
       cr = 1;
     else
-      state_maybe_utf8_putc(s, ch, is_utf8, chs, map);
+      bufi[l++] = ch;
 
     if (buf[3] == '=') break;
     c4 = base64val (buf[3]);
     ch = ((c3 & 0x3) << 6) | c4;
 
     if (cr && ch != '\n')
-      state_maybe_utf8_putc(s, ch, is_utf8, chs, map);
+      bufi[l++] = '\r';
     cr = 0;
 
     if (istext && ch == '\r')
       cr = 1;
     else
-      state_maybe_utf8_putc(s, ch, is_utf8, chs, map);
+      bufi[l++] = ch;
+    
+    if (l + 8 >= sizeof (bufi))
+      convert_to_state (cd, bufi, &l, s);
   }
+
+  if (cr) bufi[l++] = '\r';
+
+  convert_to_state (cd, bufi, &l, s);
+  convert_to_state (cd, 0, 0, s);
+
   state_reset_prefix(s);
 }
 
@@ -295,21 +365,13 @@ unsigned char decode_byte (char ch)
   return ch - 32;
 }
 
-void mutt_decode_uuencoded (STATE *s, BODY *b, int istext)
+void mutt_decode_uuencoded (STATE *s, long len, int istext, iconv_t cd)
 {
   char tmps[SHORT_STRING];
   char linelen, c, l, out;
   char *pt;
-  CHARSET_MAP *map = NULL;
-  CHARSET *chs = NULL;
-  char *charset = mutt_get_parameter("charset", b->parameter);
-  int is_utf8 = 0;
-  long len = b->length;
-  
-  if(istext && (is_utf8 = (mutt_is_utf8(charset) && !mutt_is_utf8(Charset))))
-    chs = mutt_get_charset(Charset);
-  else
-    map = mutt_get_translation(charset, Charset);
+  char bufi[BUFI_SIZE];
+  size_t k = 0;
 
   if(istext)
     state_set_prefix(s);
@@ -339,19 +401,20 @@ void mutt_decode_uuencoded (STATE *s, BODY *b, int istext)
 	out = decode_byte (*pt) << l;
 	pt++;
 	out |= (decode_byte (*pt) >> (6 - l));
-	state_maybe_utf8_putc(s, out, is_utf8, chs, map);
+	bufi[k++] = out;
 	c++;
 	if (c == linelen)
 	  break;
       }
+      convert_to_state (cd, bufi, &k, s);
       pt++;
     }
   }
+
+  convert_to_state (cd, bufi, &k, s);
+  convert_to_state (cd, 0, 0, s);
   
   state_reset_prefix(s);
-  if(is_utf8)
-    state_fput_utf8(s, '\0', chs);
-  
 }
 
 /* ----------------------------------------------------------------------------
@@ -397,6 +460,7 @@ struct enriched_state
   size_t indent_len;
   size_t word_len;
   size_t buff_used;
+  size_t param_used;
   size_t param_len;
   int tag_level[RICH_LAST_TAG];
   int WrapMargin;
@@ -529,7 +593,7 @@ static void enriched_flush (struct enriched_state *stte, int wrap)
       stte->line_max = stte->line_used;
       safe_realloc ((void **) &stte->line, stte->line_max + 1);
     }
-    strcat (stte->line, stte->buffer);
+    strcat (stte->line, stte->buffer);	/* __STRCAT_CHECKED__ */
     stte->line_len += stte->word_len;
     stte->word_len = 0;
     stte->buff_used = 0;
@@ -545,7 +609,10 @@ static void enriched_putc (int c, struct enriched_state *stte)
   {
     if (stte->tag_level[RICH_COLOR]) 
     {
-      stte->param[stte->param_len++] = c;
+      if (stte->param_used + 1 >= stte->param_len)
+	safe_realloc ((void **) &stte->param, (stte->param_len += STRING));
+
+      stte->param[stte->param_used++] = c;
     }
     return; /* nothing to do */
   }
@@ -629,7 +696,7 @@ static void enriched_set_flags (const char *tag, struct enriched_state *stte)
     tagptr++;
   
   for (i = 0, j = -1; EnrichedTags[i].tag_name; i++)
-    if (mutt_strcasecmp (EnrichedTags[i].tag_name,tagptr) == 0)
+    if (ascii_strcasecmp (EnrichedTags[i].tag_name,tagptr) == 0)
     {
       j = EnrichedTags[i].index;
       break;
@@ -646,46 +713,51 @@ static void enriched_set_flags (const char *tag, struct enriched_state *stte)
 	stte->tag_level[j]--;
       if ((stte->s->flags & M_DISPLAY) && j == RICH_PARAM && stte->tag_level[RICH_COLOR])
       {
-	stte->param[stte->param_len] = '\0';
-	if (!mutt_strcasecmp(stte->param, "black"))
+	stte->param[stte->param_used] = '\0';
+	if (!ascii_strcasecmp(stte->param, "black"))
 	{
 	  enriched_puts("\033[30m", stte);
 	}
-	else if (!mutt_strcasecmp(stte->param, "red"))
+	else if (!ascii_strcasecmp(stte->param, "red"))
 	{
 	  enriched_puts("\033[31m", stte);
 	}
-	else if (!mutt_strcasecmp(stte->param, "green"))
+	else if (!ascii_strcasecmp(stte->param, "green"))
 	{
 	  enriched_puts("\033[32m", stte);
 	}
-	else if (!mutt_strcasecmp(stte->param, "yellow"))
+	else if (!ascii_strcasecmp(stte->param, "yellow"))
 	{
 	  enriched_puts("\033[33m", stte);
 	}
-	else if (!mutt_strcasecmp(stte->param, "blue"))
+	else if (!ascii_strcasecmp(stte->param, "blue"))
 	{
 	  enriched_puts("\033[34m", stte);
 	}
-	else if (!mutt_strcasecmp(stte->param, "magenta"))
+	else if (!ascii_strcasecmp(stte->param, "magenta"))
 	{
 	  enriched_puts("\033[35m", stte);
 	}
-	else if (!mutt_strcasecmp(stte->param, "cyan"))
+	else if (!ascii_strcasecmp(stte->param, "cyan"))
 	{
 	  enriched_puts("\033[36m", stte);
 	}
-	else if (!mutt_strcasecmp(stte->param, "white"))
+	else if (!ascii_strcasecmp(stte->param, "white"))
 	{
 	  enriched_puts("\033[37m", stte);
 	}
-	stte->param_len = 0;
-	stte->param[0] = '\0';
       }
       if ((stte->s->flags & M_DISPLAY) && j == RICH_COLOR)
       {
 	enriched_puts("\033[0m", stte);
       }
+
+      /* flush parameter buffer when closing the tag */
+      if (j == RICH_PARAM)
+      {
+	stte->param_used = 0;
+	stte->param[0] = '\0';
+      }
     }
     else
       stte->tag_level[j]++;
@@ -714,6 +786,9 @@ void text_enriched_handler (BODY *a, STATE *s)
   stte.line = (char *) safe_calloc (1, stte.line_max + 1);
   stte.param = (char *) safe_calloc (1, STRING);
 
+  stte.param_len = STRING;
+  stte.param_used = 0;
+
   if (s->prefix)
   {
     state_puts (s->prefix, s);
@@ -816,6 +891,307 @@ void text_enriched_handler (BODY *a, STATE *s)
   FREE (&(stte.param));
 }                                                                              
 
+/*
+ * An implementation of RFC 2646.
+ *
+ * NOTE: This still has to be made UTF-8 aware.
+ *
+ */
+
+#define FLOWED_MAX 77
+
+static void flowed_quote (STATE *s, int level)
+{
+  int i;
+  
+  if (s->prefix)
+  {
+    if (option (OPTTEXTFLOWED))
+      level++;
+    else
+      state_puts (s->prefix, s);
+  }
+  
+  for (i = 0; i < level; i++)
+    state_putc ('>', s);
+}
+
+static int flowed_maybe_quoted (char *cont)
+{
+  return regexec ((regex_t *) QuoteRegexp.rx, cont, 0, NULL, 0) == 0;
+}
+
+static void flowed_stuff (STATE *s, char *cont, int level)
+{
+  if (!option (OPTTEXTFLOWED) && !(s->flags & M_DISPLAY))
+    return;
+
+  if (s->flags & M_DISPLAY)
+  {
+    /* 
+     * Hack: If we are in the beginning of the line and there is 
+     * some text on the line which looks like it's quoted, turn off 
+     * ANSI colors, so quote coloring doesn't affect this line. 
+     */
+    if (*cont && !level && !mutt_strcmp (Pager, "builtin") && flowed_maybe_quoted (cont))
+      state_puts ("\033[0m",s);
+  }
+  else if ((*cont == ' ') || (*cont == '>') || (!level && !mutt_strncmp (cont, "From ", 5)))
+    state_putc (' ', s);
+}
+
+static char *flowed_skip_indent (char *prefix, char *cont)
+{
+  for (; *cont == ' ' || *cont == '\t'; cont++)
+    *prefix++ = *cont;
+  *prefix = '\0';
+  return cont;
+}
+
+static int flowed_visual_strlen (char *l, int i)
+{
+  int j;
+  for (j = 0; *l; l++)
+  {
+    if (*l == '\t')
+      j += 8 - ((i + j) % 8);
+    else
+      j++;
+  }
+  
+  return j;
+}
+
+static void text_plain_flowed_handler (BODY *a, STATE *s)
+{
+  char line[LONG_STRING];
+  char indent[LONG_STRING];
+
+  int  quoted = -1;
+  int  last_quoted;
+  int  full = 1;
+  int  last_full;
+  int  col = 0, tmpcol;
+
+  int  i_add = 0;
+  int  add = 0;
+  int  soft = 0;
+  int  l, rl;
+  
+  int  flowed_max;
+  int  bytes = a->length;
+  int  actually_wrap = 0;
+  
+  char *cont = NULL;
+  char *tail = NULL;
+  char *lc = NULL;
+  char *t;
+  
+  *indent = '\0';
+  
+  if (s->prefix)
+    add = 1;
+  
+  if ((flowed_max = FLOWED_MAX) > COLS - 3)
+    flowed_max = COLS - 3;
+  if (flowed_max > COLS - WrapMargin)
+    flowed_max = COLS - WrapMargin;
+  if (flowed_max <= 0)
+    flowed_max = COLS;
+    
+
+  while (bytes > 0 && fgets (line, sizeof (line), s->fpin))
+  {
+    bytes        -= strlen (line);
+    tail          = NULL;
+
+    last_full     = full;
+    
+    /* 
+     * If the last line wasn't fully read, this is the
+     * tail of some line. 
+     */
+    actually_wrap = !last_full; 
+    
+    if ((t = strrchr (line, '\r')) || (t = strrchr (line, '\n')))
+    {
+      *t   = '\0';
+      full = 1;
+    }
+    else if ((t = strrchr (line, ' ')) || (t = strrchr (line, '\t')))
+    {
+      /* 
+       * Bad: We have a line of more than LONG_STRING characters.
+       * (Which SHOULD NOT happen, since lines SHOULD be <= 79
+       * characters long.)
+       * 
+       * Try to simulate a soft line break at a word boundary.
+       * Handle the rest of the line next time.
+       * 
+       * Give up when we have a single word which is longer than
+       * LONG_STRING characters.  It will just be split into parts,
+       * with a hard line break in between. 
+       */
+
+      full = 0;
+      l    = strlen (t + 1);
+      t[0] = ' ';
+      t[1] = '\0';
+
+      if (l)
+      {
+	fseek (s->fpin, -l, SEEK_CUR);
+	bytes += l;
+      }
+    }
+    else
+      full = 0;
+
+    last_quoted = quoted;
+
+    if (last_full)
+    {
+      /* 
+       * We are in the beginning of a new line. Determine quote level
+       * and indentation prefix 
+       */
+      for (quoted = 0; line[quoted] == '>'; quoted++)
+	;
+      
+      cont = line + quoted;
+      
+      /* undo space stuffing */
+      if (*cont == ' ')
+	cont++;
+
+      /* If there is an indentation, record it. */
+      cont  = flowed_skip_indent (indent, cont);
+      i_add = flowed_visual_strlen (indent, quoted + add);
+    }
+    else
+    {
+      /* 
+       * This is just the tail of some over-long line. Keep
+       * indentation and quote levels.  Don't unstuff.
+       */
+      cont = line;
+    }
+
+    /* If we have a change in quoting depth, wrap. */
+
+    if (col && last_quoted != quoted && last_quoted >= 0)
+    {
+      state_putc ('\n', s);
+      col = 0;
+    }
+    
+    do 
+    {
+      if (tail)
+	cont = tail;
+
+      SKIPWS (cont);
+      
+      tail = NULL;
+      soft = 0;
+      
+      /* try to find a point for word wrapping */
+
+    retry_wrap:
+      l  = flowed_visual_strlen (cont, quoted + i_add + add + col);
+      rl = mutt_strlen (cont);
+      if (quoted + i_add + add + col + l > flowed_max)
+      {
+	actually_wrap = 1;
+
+	for (tmpcol = quoted + i_add + add + col, t = cont;
+	     *t && tmpcol < flowed_max; t++)
+	{
+	  if (*t == ' ' || *t == '\t')
+	    tail = t;
+	  if (*t == '\t')
+	    tmpcol = (tmpcol & ~7) + 8;
+	  else
+	    tmpcol++;
+	}
+	
+	if (tail)
+	{
+	  *tail++ = '\0';
+	  soft = 2;
+	}
+      }
+
+      /* We seem to be desperate.  Get me a new line, and retry. */
+      if (!tail && (quoted + add + col + i_add + l > flowed_max) && col)
+      {
+	state_putc ('\n', s);
+	col = 0;
+	goto retry_wrap;
+      }
+
+      /* Detect soft line breaks. */
+      if (!soft && ascii_strcmp (cont, "-- "))
+      {
+	lc = strrchr (cont, ' ');
+	if (lc && lc[1] == '\0')
+	  soft = 1;
+      }
+
+      /* 
+       * If we are in the beginning of an output line, do quoting
+       * and stuffing. 
+       * 
+       * We have to temporarily assemble the line since display
+       * stuffing (i.e., turning off quote coloring) may depend on
+       * the line's actual content.  You never know what people put
+       * into their regular expressions. 
+       */
+      if (!col)
+      {
+	char tmp[LONG_STRING];
+	snprintf (tmp, sizeof (tmp), "%s%s", indent, cont);
+
+	flowed_quote (s, quoted);
+	flowed_stuff (s, tmp, quoted + add);
+
+	state_puts (indent, s);
+      }
+
+      /* output the text */
+      state_puts (cont, s);
+      col += flowed_visual_strlen (cont, quoted + i_add + add + col);
+      
+      /* possibly indicate a soft line break */
+      if (soft == 2)
+      {
+	state_putc (' ', s);
+	col++;
+      }
+      
+      /* 
+       * Wrap if this display line corresponds to a 
+       * text line. Don't wrap if we changed the line.
+       */
+      if (!soft || (!actually_wrap && full))
+      {
+	state_putc ('\n', s);
+	col = 0;
+      }
+    }
+    while (tail);
+  }
+
+  if (col)
+    state_putc ('\n', s);
+  
+}
+
+
+
+
+
+
 #define TXTHTML     1
 #define TXTPLAIN    2
 #define TXTENRICHED 3
@@ -855,10 +1231,10 @@ static void alternative_handler (BODY *a, STATE *s)
     while (b)
     {
       const char *bt = TYPE(b);
-      if (!mutt_strncasecmp (bt, t->data, btlen) && bt[btlen] == 0)
+      if (!ascii_strncasecmp (bt, t->data, btlen) && bt[btlen] == 0)
       {
 	/* the basetype matches */
-	if (wild || !mutt_strcasecmp (t->data + btlen + 1, b->subtype))
+	if (wild || !ascii_strcasecmp (t->data + btlen + 1, b->subtype))
 	{
 	  choice = b;
 	}
@@ -903,17 +1279,17 @@ static void alternative_handler (BODY *a, STATE *s)
     {
       if (b->type == TYPETEXT)
       {
-	if (! mutt_strcasecmp ("plain", b->subtype) && type <= TXTPLAIN)
+	if (! ascii_strcasecmp ("plain", b->subtype) && type <= TXTPLAIN)
 	{
 	  choice = b;
 	  type = TXTPLAIN;
 	}
-	else if (! mutt_strcasecmp ("enriched", b->subtype) && type <= TXTENRICHED)
+	else if (! ascii_strcasecmp ("enriched", b->subtype) && type <= TXTENRICHED)
 	{
 	  choice = b;
 	  type = TXTENRICHED;
 	}
-	else if (! mutt_strcasecmp ("html", b->subtype) && type <= TXTHTML)
+	else if (! ascii_strcasecmp ("html", b->subtype) && type <= TXTHTML)
 	{
 	  choice = b;
 	  type = TXTHTML;
@@ -950,6 +1326,7 @@ static void alternative_handler (BODY *a, STATE *s)
   else if (s->flags & M_DISPLAY)
   {
     /* didn't find anything that we could display! */
+    state_mark_attach (s);
     state_puts(_("[-- Error:  Could not display any parts of Multipart/Alternative! --]\n"), s);
   }
 }
@@ -976,7 +1353,7 @@ void message_handler (BODY *a, STATE *s)
   if (b->parts)
   {
     mutt_copy_hdr (s->fpin, s->fpout, off_start, b->parts->offset,
-	(((s->flags & M_WEED) || ((s->flags & M_DISPLAY) && option (OPTWEED))) ? (CH_WEED | CH_REORDER) : 0) |
+	(((s->flags & M_WEED) || ((s->flags & (M_DISPLAY|M_PRINTING)) && option (OPTWEED))) ? (CH_WEED | CH_REORDER) : 0) |
 	(s->prefix ? CH_PREFIX : 0) | CH_DECODE | CH_FROM, s->prefix);
 
     if (s->prefix)
@@ -1008,9 +1385,9 @@ int mutt_can_decode (BODY *a)
 
 
 
-#ifdef _PGPPATH
-    if (mutt_strcasecmp (a->subtype, "signed") == 0 ||
-	mutt_strcasecmp (a->subtype, "encrypted") == 0)
+#ifdef HAVE_PGP
+    if (ascii_strcasecmp (a->subtype, "signed") == 0 ||
+	ascii_strcasecmp (a->subtype, "encrypted") == 0)
       return (1);
     else
 #endif
@@ -1030,7 +1407,7 @@ int mutt_can_decode (BODY *a)
 
 
 
-#ifdef _PGPPATH
+#ifdef HAVE_PGP
   else if (a->type == TYPEAPPLICATION)
   {
     if (mutt_is_application_pgp(a))
@@ -1058,7 +1435,7 @@ void multipart_handler (BODY *a, STATE *s)
     b->length = (long) st.st_size;
     b->parts = mutt_parse_multipart (s->fpin,
 		  mutt_get_parameter ("boundary", a->parameter),
-		  (long) st.st_size, mutt_strcasecmp ("digest", a->subtype) == 0);
+		  (long) st.st_size, ascii_strcasecmp ("digest", a->subtype) == 0);
   }
   else
     b = a;
@@ -1067,6 +1444,7 @@ void multipart_handler (BODY *a, STATE *s)
   {
     if (s->flags & M_DISPLAY)
     {
+      state_mark_attach (s);
       state_printf (s, _("[-- Attachment #%d"), count);
       if (p->description || p->filename || p->form_name)
       {
@@ -1078,6 +1456,7 @@ void multipart_handler (BODY *a, STATE *s)
 
       mutt_pretty_size (length, sizeof (length), p->length);
       
+      state_mark_attach (s);
       state_printf (s, _("[-- Type: %s/%s, Encoding: %s, Size: %s --]\n"),
 		    TYPE (p), p->subtype, ENCODING (p->encoding), length);
       if (!option (OPTWEED))
@@ -1137,6 +1516,7 @@ void autoview_handler (BODY *a, STATE *s)
 
     if (s->flags & M_DISPLAY)
     {
+      state_mark_attach (s);
       state_printf (s, _("[-- Autoview using %s --]\n"), command);
       mutt_message(_("Invoking autoview command: %s"),command);
     }
@@ -1152,18 +1532,29 @@ void autoview_handler (BODY *a, STATE *s)
 
     if(!piped)
     {
-      fclose (fpin);
+      safe_fclose (&fpin);
       thepid = mutt_create_filter (command, NULL, &fpout, &fperr);
     }
     else
     {
-      unlink(tempfile);
-      fflush(fpin);
-      rewind(fpin);
+      unlink (tempfile);
+      fflush (fpin);
+      rewind (fpin);
       thepid = mutt_create_filter_fd (command, NULL, &fpout, &fperr,
 				      fileno(fpin), -1, -1);
     }
 
+    if (thepid < 0)
+    {
+      mutt_perror _("Can't create filter");
+      if (s->flags & M_DISPLAY)
+      {
+	state_mark_attach (s);
+	state_printf (s, _("[-- Can't run %s. --]\n"), command);
+      }
+      goto bail;
+    }
+    
     if (s->prefix)
     {
       while (fgets (buffer, sizeof(buffer), fpout) != NULL)
@@ -1174,8 +1565,11 @@ void autoview_handler (BODY *a, STATE *s)
       /* check for data on stderr */
       if (fgets (buffer, sizeof(buffer), fperr)) 
       {
-	if (s->flags & M_DISPLAY) 
+	if (s->flags & M_DISPLAY)
+	{
+	  state_mark_attach (s);
 	  state_printf (s, _("[-- Autoview stderr of %s --]\n"), command);
+	}
 
 	state_puts (s->prefix, s);
 	state_puts (buffer, s);
@@ -1193,19 +1587,24 @@ void autoview_handler (BODY *a, STATE *s)
       if (fgets (buffer, sizeof(buffer), fperr))
       {
 	if (s->flags & M_DISPLAY)
+	{
+	  state_mark_attach (s);
 	  state_printf (s, _("[-- Autoview stderr of %s --]\n"), 
 			command);
+	}
 	
 	state_puts (buffer, s);
 	mutt_copy_stream (fperr, s->fpout);
       }
     }
 
-    fclose (fpout);
-    fclose (fperr);
+  bail:
+    safe_fclose (&fpout);
+    safe_fclose (&fperr);
+
     mutt_wait_filter (thepid);
-    if(piped)
-      fclose(fpin);
+    if (piped)
+      safe_fclose (&fpin);
     else
       mutt_unlink (tempfile);
 
@@ -1225,7 +1624,10 @@ static void external_body_handler (BODY *b, STATE *s)
   if (!access_type)
   {
     if (s->flags & M_DISPLAY)
+    {
+      state_mark_attach (s);
       state_puts (_("[-- Error: message/external-body has no access-type parameter --]\n"), s);
+    }
     return;
   }
 
@@ -1235,13 +1637,14 @@ static void external_body_handler (BODY *b, STATE *s)
   else
     expire = -1;
 
-  if (!mutt_strcasecmp (access_type, "x-mutt-deleted"))
+  if (!ascii_strcasecmp (access_type, "x-mutt-deleted"))
   {
-    if (s->flags & M_DISPLAY)
+    if (s->flags & (M_DISPLAY|M_PRINTING))
     {
       char *length;
       char pretty_size[10];
-
+      
+      state_mark_attach (s);
       state_printf (s, _("[-- This %s/%s attachment "),
 	       TYPE(b->parts), b->parts->subtype);
       length = mutt_get_parameter ("length", b->parameter);
@@ -1254,20 +1657,31 @@ static void external_body_handler (BODY *b, STATE *s)
       state_puts (_("has been deleted --]\n"), s);
 
       if (expire != -1)
+      {
+	state_mark_attach (s);
 	state_printf (s, _("[-- on %s --]\n"), expiration);
+      }
+      if (b->parts->filename)
+      {
+	state_mark_attach (s);
+	state_printf (s, _("[-- name: %s --]\n"), b->parts->filename);
+      }
+
       mutt_copy_hdr (s->fpin, s->fpout, ftell (s->fpin), b->parts->offset,
 		     (option (OPTWEED) ? (CH_WEED | CH_REORDER) : 0) |
 		     CH_DECODE , NULL);
     }
   }
-  else if(expire < time(NULL))
+  else if(expiration && expire < time(NULL))
   {
     if (s->flags & M_DISPLAY)
     {
-      state_printf (s, _("[-- This %s/%s attachment is not included, --]\n"
-			 "[-- and the indicated external source has --]\n"
-			 "[-- expired. --]\n"),
+      state_mark_attach (s);
+      state_printf (s, _("[-- This %s/%s attachment is not included, --]\n"),
 		    TYPE(b->parts), b->parts->subtype);
+      state_attach_puts (_("[-- and the indicated external source has --]\n"
+			   "[-- expired. --]\n"), s);
+
       mutt_copy_hdr(s->fpin, s->fpout, ftell (s->fpin), b->parts->offset,
 		    (option (OPTWEED) ? (CH_WEED | CH_REORDER) : 0) |
 		    CH_DECODE, NULL);
@@ -1277,10 +1691,14 @@ static void external_body_handler (BODY *b, STATE *s)
   {
     if (s->flags & M_DISPLAY)
     {
+      state_mark_attach (s);
       state_printf (s,
-	       _("[-- This %s/%s attachment is not included, --]\n"
-		 "[-- and the indicated access-type %s is unsupported --]\n"),
-	       TYPE(b->parts), b->parts->subtype, access_type);
+		    _("[-- This %s/%s attachment is not included, --]\n"),
+		    TYPE (b->parts), b->parts->subtype);
+      state_mark_attach (s);
+      state_printf (s, 
+		    _("[-- and the indicated access-type %s is unsupported --]\n"),
+		    access_type);
       mutt_copy_hdr (s->fpin, s->fpout, ftell (s->fpin), b->parts->offset,
 		     (option (OPTWEED) ? (CH_WEED | CH_REORDER) : 0) |
 		     CH_DECODE , NULL);
@@ -1290,22 +1708,35 @@ static void external_body_handler (BODY *b, STATE *s)
 
 void mutt_decode_attachment (BODY *b, STATE *s)
 {
+  int istext = mutt_is_text_type (b->type, b->subtype);
+  iconv_t cd = (iconv_t)(-1);
+
+  if (istext && s->flags & M_CHARCONV)
+  {
+    char *charset = mutt_get_parameter ("charset", b->parameter);
+    if (charset && Charset)
+      cd = mutt_iconv_open (Charset, charset, M_ICONV_HOOK_FROM);
+  }
+
   fseek (s->fpin, b->offset, 0);
   switch (b->encoding)
   {
     case ENCQUOTEDPRINTABLE:
-      mutt_decode_quoted (s, b, mutt_is_text_type (b->type, b->subtype));
+      mutt_decode_quoted (s, b->length, istext, cd);
       break;
     case ENCBASE64:
-      mutt_decode_base64 (s, b, mutt_is_text_type (b->type, b->subtype));
+      mutt_decode_base64 (s, b->length, istext, cd);
       break;
     case ENCUUENCODED:
-      mutt_decode_uuencoded (s, b, mutt_is_text_type (b->type, b->subtype));
+      mutt_decode_uuencoded (s, b->length, istext, cd);
       break;
     default:
-      mutt_decode_xbit (s, b, mutt_is_text_type (b->type, b->subtype));
+      mutt_decode_xbit (s, b->length, istext, cd);
       break;
   }
+
+  if (cd != (iconv_t)(-1))
+    iconv_close (cd);
 }
 
 void mutt_body_handler (BODY *b, STATE *s)
@@ -1319,6 +1750,8 @@ void mutt_body_handler (BODY *b, STATE *s)
   size_t tmplength = 0;
   char type[STRING];
 
+  int oflags = s->flags;
+  
   /* first determine which handler to use to process this part */
 
   snprintf (type, sizeof (type), "%s/%s", TYPE (b), b->subtype);
@@ -1327,30 +1760,36 @@ void mutt_body_handler (BODY *b, STATE *s)
     rfc1524_entry *entry = rfc1524_new_entry ();
 
     if (rfc1524_mailcap_lookup (b, type, entry, M_AUTOVIEW))
-      handler = autoview_handler;
+    {
+      handler   = autoview_handler;
+      s->flags &= ~M_CHARCONV;
+    }
     rfc1524_free_entry (&entry);
   }
   else if (b->type == TYPETEXT)
   {
-    if (mutt_strcasecmp ("plain", b->subtype) == 0)
+    if (ascii_strcasecmp ("plain", b->subtype) == 0)
     {
       /* avoid copying this part twice since removing the transfer-encoding is
        * the only operation needed.
        */
-      plaintext = 1;
+      if (ascii_strcasecmp ("flowed", mutt_get_parameter ("format", b->parameter)) == 0)
+	handler = text_plain_flowed_handler;
+      else
+	plaintext = 1;
     }
-    else if (mutt_strcasecmp ("enriched", b->subtype) == 0)
+    else if (ascii_strcasecmp ("enriched", b->subtype) == 0)
       handler = text_enriched_handler;
-    else if (mutt_strcasecmp ("rfc822-headers", b->subtype) == 0)
+    else if (ascii_strcasecmp ("rfc822-headers", b->subtype) == 0)
       plaintext = 1;
   }
   else if (b->type == TYPEMESSAGE)
   {
     if(mutt_is_message_type(b->type, b->subtype))
       handler = message_handler;
-    else if (!mutt_strcasecmp ("delivery-status", b->subtype))
+    else if (!ascii_strcasecmp ("delivery-status", b->subtype))
       plaintext = 1;
-    else if (!mutt_strcasecmp ("external-body", b->subtype))
+    else if (!ascii_strcasecmp ("external-body", b->subtype))
       handler = external_body_handler;
   }
   else if (b->type == TYPEMULTIPART)
@@ -1358,40 +1797,41 @@ void mutt_body_handler (BODY *b, STATE *s)
 
 
 
-#ifdef _PGPPATH
+#ifdef HAVE_PGP
     char *p;
-#endif /* _PGPPATH */
+#endif /* HAVE_PGP */
 
 
 
-    if (mutt_strcasecmp ("alternative", b->subtype) == 0)
+    if (ascii_strcasecmp ("alternative", b->subtype) == 0)
       handler = alternative_handler;
 
 
 
-#ifdef _PGPPATH
-    else if (mutt_strcasecmp ("signed", b->subtype) == 0)
+#ifdef HAVE_PGP
+    else if (ascii_strcasecmp ("signed", b->subtype) == 0)
     {
       p = mutt_get_parameter ("protocol", b->parameter);
 
       if (!p)
         mutt_error _("Error: multipart/signed has no protocol.");
-      else if (mutt_strcasecmp ("application/pgp-signature", p) == 0)
+      else if (ascii_strcasecmp ("application/pgp-signature", p) == 0 ||
+	       ascii_strcasecmp ("multipart/mixed", p) == 0)
       {
 	if (s->flags & M_VERIFY)
 	  handler = pgp_signed_handler;
       }
     }
-    else if (mutt_strcasecmp ("encrypted", b->subtype) == 0)
+    else if (ascii_strcasecmp ("encrypted", b->subtype) == 0)
     {
       p = mutt_get_parameter ("protocol", b->parameter);
 
       if (!p)
         mutt_error _("Error: multipart/encrypted has no protocol parameter!");
-      else if (mutt_strcasecmp ("application/pgp-encrypted", p) == 0)
+      else if (ascii_strcasecmp ("application/pgp-encrypted", p) == 0)
         handler = pgp_encrypted_handler;
     }
-#endif /* _PGPPATH */
+#endif /* HAVE_PGP */
 
 
 
@@ -1401,13 +1841,13 @@ void mutt_body_handler (BODY *b, STATE *s)
 
 
 
-#ifdef _PGPPATH
+#ifdef HAVE_PGP
   else if (b->type == TYPEAPPLICATION)
   {
     if (mutt_is_application_pgp(b))
-      handler = application_pgp_handler;
+      handler = pgp_application_pgp_handler;
   }
-#endif /* _PGPPATH */
+#endif /* HAVE_PGP */
 
 
 
@@ -1417,7 +1857,12 @@ void mutt_body_handler (BODY *b, STATE *s)
 
     /* see if we need to decode this part before processing it */
     if (b->encoding == ENCBASE64 || b->encoding == ENCQUOTEDPRINTABLE ||
-	b->encoding == ENCUUENCODED || plaintext)
+	b->encoding == ENCUUENCODED || plaintext || 
+	mutt_is_text_type (b->type, b->subtype))	/* text subtypes may
+							 * require character
+							 * set conversion even
+							 * with 8bit encoding.
+							 */
     {
       int origType = b->type;
       char *savePrefix = NULL;
@@ -1430,7 +1875,7 @@ void mutt_body_handler (BODY *b, STATE *s)
 	if ((s->fpout = safe_fopen (tempfile, "w")) == NULL)
 	{
 	  mutt_error _("Unable to open temporary file!");
-	  return;
+	  goto bail;
 	}
 	/* decoding the attachment changes the size and offset, so save a copy
 	 * of the "real" values now, and restore them after processing
@@ -1488,6 +1933,7 @@ void mutt_body_handler (BODY *b, STATE *s)
   }
   else if (s->flags & M_DISPLAY)
   {
+    state_mark_attach (s);
     state_printf (s, _("[-- %s/%s is unsupported "), TYPE (b), b->subtype);
     if (!option (OPTVIEWATTACH))
     {
@@ -1499,4 +1945,7 @@ void mutt_body_handler (BODY *b, STATE *s)
     }
     fputs (" --]\n", s->fpout);
   }
+  
+  bail:
+  s->flags = oflags;
 }
