@@ -19,22 +19,36 @@
 #include "config.h"
 
 #include <stdio.h>
+#include <stdlib.h>
 #ifdef HAVE_UNISTD_H
 # include <unistd.h> /* needed for SEEK_SET */
 #endif
+#ifdef HAVE_UNIX_H
+# include <unix.h>   /* needed for snprintf on QNX. */
+#endif
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <time.h>
 #include <limits.h>
 #include <stdarg.h>
 #include <signal.h>
+#ifdef HAVE_WCHAR_H
+# include <wchar.h>
+#endif
+#if defined(HAVE_WCTYPE_H) && defined(HAVE_WC_FUNCS)
+# include <wctype.h>
+#endif
 
 #ifndef _POSIX_PATH_MAX
 #include <posix1_lim.h>
 #endif
 
+#include <pwd.h>
+#include <grp.h>
+
 #include "rfc822.h"
 #include "hash.h"
+#include "charset.h"
 
 #ifdef SUBVERSION
 # define MUTT_VERSION (VERSION SUBVERSION)
@@ -123,9 +137,12 @@ typedef enum
 #define M_FCCHOOK	(1<<3)
 #define M_SAVEHOOK	(1<<4)
 #define M_CHARSETHOOK	(1<<5)
+#define M_ICONVHOOK 	(1<<6)
+#define M_MESSAGEHOOK	(1<<7)
 #ifdef HAVE_PGP
-#define M_PGPHOOK	(1<<6)
+#define M_PGPHOOK	(1<<8)
 #endif
+#define M_ACCOUNTHOOK	(1<<9)
 
 /* tree characters for linearize_tree and print_enriched_string */
 #define M_TREE_LLCORNER		1
@@ -137,7 +154,11 @@ typedef enum
 #define M_TREE_RARROW		7
 #define M_TREE_STAR		8
 #define M_TREE_HIDDEN		9
-#define M_TREE_MAX		10
+#define M_TREE_EQUALS		10
+#define M_TREE_TTEE		11
+#define M_TREE_BTEE		12
+#define M_TREE_MISSING		13
+#define M_TREE_MAX		14
 
 #define M_THREAD_COLLAPSE	(1<<0)
 #define M_THREAD_UNCOLLAPSE	(1<<1)
@@ -180,6 +201,7 @@ enum
   M_FROM,
   M_DATE,
   M_DATE_RECEIVED,
+  M_DUPLICATED,
   M_ID,
   M_BODY,
   M_HEADER,
@@ -199,6 +221,7 @@ enum
   M_PGP_ENCRYPT,
   M_PGP_KEY,
 #endif
+  M_XLABEL,
   
   /* Options for Mailcap lookup */
   M_EDIT,
@@ -235,6 +258,10 @@ enum
   OPT_PGPTRADITIONAL, /* create old-style PGP messages */
 #endif
 
+#ifdef USE_SSL
+  OPT_SSLSTARTTLS,
+#endif
+
   OPT_PRINT,
   OPT_INCLUDE,
   OPT_DELETE,
@@ -242,6 +269,10 @@ enum
   OPT_MIMEFWD,
   OPT_MOVE,
   OPT_COPY,
+#ifdef USE_POP
+  OPT_POPDELETE,
+  OPT_POPRECONNECT,
+#endif
   OPT_POSTPONE,
   OPT_QUIT,
   OPT_REPLYTO,
@@ -263,10 +294,17 @@ enum
 #define SENDKEY		(1<<7)
 #define SENDRESEND	(1<<8)
 
+/* flags to _mutt_select_file() */
+#define M_SEL_BUFFY	(1<<0)
+#define M_SEL_MULTI	(1<<1)
+#define M_SEL_FOLDER	(1<<2)
+
+
 /* boolean vars */
 enum
 {
   OPTALLOW8BIT,
+  OPTALLOWANSI,
   OPTARROWCURSOR,
   OPTASCIICHARS,
   OPTASKBCC,
@@ -282,6 +320,8 @@ enum
   OPTCONFIRMAPPEND,
   OPTCONFIRMCREATE,
   OPTDELETEUNTAG,
+  OPTDIGESTCOLLAPSE,
+  OPTDUPTHREADS,
   OPTEDITHDRS,
   OPTENCODEFROM,
   OPTENVFROM,
@@ -296,23 +336,30 @@ enum
   OPTHEADER,
   OPTHELP,
   OPTHIDDENHOST,
+  OPTHIDELIMITED,
+  OPTHIDEMISSING,
+  OPTHIDETOPLIMITED,
+  OPTHIDETOPMISSING,
   OPTIGNORELISTREPLYTO,
 #ifdef USE_IMAP
   OPTIMAPLSUB,
   OPTIMAPPASSIVE,
+  OPTIMAPPEEK,
   OPTIMAPSERVERNOISE,
 # ifdef USE_SSL
   OPTIMAPFORCESSL,
 # endif
 #endif
-#ifdef USE_SSL
+#if defined(USE_SSL) || defined(USE_NSS)
   OPTSSLV2,
   OPTSSLV3,
   OPTTLSV1,
   OPTSSLSYSTEMCERTS,
 #endif
   OPTIMPLICITAUTOVIEW,
+  OPTKEEPFLAGGED,
   OPTMAILCAPSANITIZE,
+  OPTMAILDIRTRASH,
   OPTMARKERS,
   OPTMARKOLD,
   OPTMENUSCROLL,	/* scroll menu instead of implicit next-page */
@@ -323,9 +370,12 @@ enum
   OPTPAGERSTOP,
   OPTPIPEDECODE,
   OPTPIPESPLIT,
-  OPTPOPDELETE,
+#ifdef USE_POP
+  OPTPOPAUTHTRYALL,
   OPTPOPLAST,
+#endif
   OPTPRINTDECODE,
+  OPTPRINTSPLIT,
   OPTPROMPTAFTER,
   OPTREADONLY,
   OPTREPLYSELF,
@@ -339,16 +389,22 @@ enum
   OPTSAVENAME,
   OPTSCORE,
   OPTSIGDASHES,
+  OPTSIGONTOP,
   OPTSORTRE,
   OPTSTATUSONTOP,
   OPTSTRICTTHREADS,
   OPTSUSPEND,
+  OPTTEXTFLOWED,
   OPTTHOROUGHSRC,
+  OPTTHREADRECEIVED,
   OPTTILDE,
   OPTUNCOLLAPSEJUMP,
   OPTUSE8BITMIME,
   OPTUSEDOMAIN,
   OPTUSEFROM,
+#ifdef HAVE_GETADDRINFO
+  OPTUSEIPV6,
+#endif
   OPTWAITKEY,
   OPTWEED,
   OPTWRAP,
@@ -361,6 +417,7 @@ enum
 #ifdef HAVE_PGP
   OPTPGPAUTOSIGN,
   OPTPGPAUTOENCRYPT,
+  OPTPGPIGNORESUB,
   OPTPGPLONGIDS,
   OPTPGPREPLYENCRYPT,
   OPTPGPREPLYSIGN,
@@ -387,6 +444,7 @@ enum
   OPTSIGNALSBLOCKED,	/* (pseudo) using by mutt_block_signals () */
   OPTSYSSIGNALSBLOCKED,	/* (pseudo) using by mutt_block_signals_system () */
   OPTNEEDRESORT,	/* (pseudo) used to force a re-sort */
+  OPTRESORTINIT,	/* (pseudo) used to force the next resort to be from scratch */
   OPTVIEWATTACH,	/* (pseudo) signals that we are viewing attachments */
   OPTFORCEREDRAWINDEX,	/* (pseudo) used to force a redraw in the main index */
   OPTFORCEREDRAWPAGER,	/* (pseudo) used to force a redraw in the pager */
@@ -397,7 +455,7 @@ enum
 			 * 	    functions while we are executing an
 			 * 	    external program.
 			 */
-  
+  OPTMENUCALLER,	/* (pseudo) tell menu to give caller a take */
 #ifdef HAVE_PGP
   OPTPGPCHECKTRUST,	/* (pseudo) used by pgp_select_key () */
   OPTDONTHANDLEPGPKEYS,	/* (pseudo) used to extract PGP keys */
@@ -441,10 +499,12 @@ void mutt_init (int, LIST *);
 
 typedef struct alias
 {
+  struct alias *self;		/* XXX - ugly hack */
   char *name;
   ADDRESS *addr;
   struct alias *next;
   short tagged;
+  short del;
   short num;
 } ALIAS;
 
@@ -463,7 +523,9 @@ typedef struct envelope
   char *message_id;
   char *supersedes;
   char *date;
+  char *x_label;
   LIST *references;		/* message references (in reverse order) */
+  LIST *in_reply_to;		/* in-reply-to header content */
   LIST *userhdrs;		/* user defined headers */
 } ENVELOPE;
 
@@ -479,6 +541,7 @@ typedef struct content
 {
   long hibin;              /* 8-bit characters */
   long lobin;              /* unprintable 7-bit chars (eg., control chars) */
+  long crlf;		   /* '\r' and '\n' characters */
   long ascii;              /* number of ascii chars */
   long linemax;            /* length of the longest line in the file */
   unsigned int space : 1;  /* whitespace at the end of lines? */
@@ -533,12 +596,19 @@ typedef struct body
 				 */
   unsigned int tagged : 1;
   unsigned int deleted : 1;	/* attachment marked for deletion */
+
   unsigned int noconv : 1;	/* don't do character set conversion */
+  unsigned int force_charset : 1; 
+  				/* send mode: don't adjust the character
+				 * set when in send-mode.
+				 */
 
 #ifdef HAVE_PGP
   unsigned int goodsig : 1;	/* good PGP signature */
 #endif
 
+  unsigned int collapsed : 1;	/* used by recvattach */
+
 } BODY;
 
 typedef struct header
@@ -547,23 +617,26 @@ typedef struct header
   unsigned int pgp : 4;
 #endif
 
-  unsigned int mime : 1;    /* has a Mime-Version header? */
-  unsigned int flagged : 1; /* marked important? */
+  unsigned int mime : 1;    		/* has a Mime-Version header? */
+  unsigned int flagged : 1; 		/* marked important? */
   unsigned int tagged : 1;
   unsigned int deleted : 1;
   unsigned int changed : 1;
-  unsigned int attach_del : 1; /* has an attachment marked for deletion */
+  unsigned int attach_del : 1; 		/* has an attachment marked for deletion */
   unsigned int old : 1;
   unsigned int read : 1;
-  unsigned int expired : 1; /* already expired? */
-  unsigned int superseded : 1; /* got superseded? */
+  unsigned int expired : 1; 		/* already expired? */
+  unsigned int superseded : 1; 		/* got superseded? */
   unsigned int replied : 1;
-  unsigned int subject_changed : 1; /* used for threading */
-  unsigned int display_subject : 1; /* used for threading */
-  unsigned int fake_thread : 1;     /* no ref matched, but subject did */
-  unsigned int threaded : 1;        /* message has been threaded */
-  unsigned int recip_valid : 1;  /* is_recipient is valid */
-  unsigned int active : 1;	    /* message is not to be removed */
+  unsigned int subject_changed : 1; 	/* used for threading */
+  unsigned int threaded : 1;	    	/* used for threading */
+  unsigned int display_subject : 1; 	/* used for threading */
+  unsigned int recip_valid : 1;  	/* is_recipient is valid */
+  unsigned int active : 1;	    	/* message is not to be removed */
+  unsigned int trash : 1;		/* message is marked as trashed on disk.
+					 * This flag is used by the maildir_trash
+					 * option.
+					 */
   
   /* timezone of the sender of this message */
   unsigned int zhours : 5;
@@ -575,43 +648,60 @@ typedef struct header
   unsigned int matched : 1;
 
   /* the following are used to support collapsing threads  */
-  unsigned int collapsed : 1; /* is this message part of a collapsed thread? */
-  unsigned int limited : 1;   /* is this message in a limited view?  */
-  size_t num_hidden;          /* number of hidden messages in this view */
+  unsigned int collapsed : 1; 	/* is this message part of a collapsed thread? */
+  unsigned int limited : 1;   	/* is this message in a limited view?  */
+  size_t num_hidden;          	/* number of hidden messages in this view */
 
-  short recipient;	/* user_is_recipient()'s return value, cached */
+  short recipient;		/* user_is_recipient()'s return value, cached */
   
-  int pair; /* color-pair to use when displaying in the index */
+  int pair; 			/* color-pair to use when displaying in the index */
 
-  time_t date_sent;     /* time when the message was sent (UTC) */
-  time_t received;      /* time when the message was placed in the mailbox */
-  long offset;          /* where in the stream does this message begin? */
-  int lines;		/* how many lines in the body of this message? */
-  int index;		/* the absolute (unsorted) message number */
-  int msgno;		/* number displayed to the user */
-  int virtual;		/* virtual message number */
+  time_t date_sent;     	/* time when the message was sent (UTC) */
+  time_t received;      	/* time when the message was placed in the mailbox */
+  long offset;          	/* where in the stream does this message begin? */
+  int lines;			/* how many lines in the body of this message? */
+  int index;			/* the absolute (unsorted) message number */
+  int msgno;			/* number displayed to the user */
+  int virtual;			/* virtual message number */
   int score;
-  ENVELOPE *env;	/* envelope information */
-  BODY *content;	/* list of MIME parts */
+  ENVELOPE *env;		/* envelope information */
+  BODY *content;		/* list of MIME parts */
   char *path;
   
-  /* the following are used for threading support */
-  struct header *parent;
-  struct header *child;  /* decendants of this message */
-  struct header *next;   /* next message in this thread */
-  struct header *prev;   /* previous message in thread */
-  struct header *last_sort; /* last message in subthread, for secondary SORT_LAST */
-  char *tree;            /* character string to print thread tree */
+  char *tree;           	/* character string to print thread tree */
+  struct thread *thread;
 
 #ifdef MIXMASTER
   LIST *chain;
 #endif
 
-#ifdef USE_IMAP
-  void *data;            /* driver-specific data (only used by IMAP */
+#ifdef USE_POP
+  int refno;			/* message number on server */
+#endif
+
+#if defined USE_POP || defined USE_IMAP
+  void *data;            	/* driver-specific data */
 #endif
 } HEADER;
 
+typedef struct thread
+{
+  unsigned int fake_thread : 1;
+  unsigned int duplicate_thread : 1;
+  unsigned int sort_children : 1;
+  unsigned int check_subject : 1;
+  unsigned int visible : 1;
+  unsigned int deep : 1;
+  unsigned int subtree_visible : 2;
+  unsigned int next_subtree_visible : 1;
+  struct thread *parent;
+  struct thread *child;
+  struct thread *next;
+  struct thread *prev;
+  HEADER *message;
+  HEADER *sort_key;
+} THREAD;
+
 #include "mutt_regex.h"
 
 /* flag to mutt_pattern_comp() */
@@ -644,9 +734,10 @@ typedef struct
   char *pattern;                /* limit pattern string */
   pattern_t *limit_pattern;     /* compiled limit pattern */
   HEADER **hdrs;
-  HEADER *tree;			/* top of thread tree */
+  THREAD *tree;			/* top of thread tree */
   HASH *id_hash;		/* hash table by msg id */
   HASH *subj_hash;		/* hash table by subject */
+  HASH *thread_hash;		/* hash table for threading */
   int *v2r;			/* mapping from virtual to real msgno */
   int hdrmax;			/* number of pointers in hdrs */
   int msgcount;			/* number of messages in the mailbox */
@@ -657,7 +748,7 @@ typedef struct
   int deleted;			/* how many deleted messages */
   int flagged;			/* how many flagged messages */
   int msgnotreadyet;		/* which msg "new" in pager, -1 if none */
-#ifdef USE_IMAP
+#if defined USE_POP || defined USE_IMAP
   void *data;			/* driver specific data */
 #endif /* USE_IMAP */
 
@@ -690,6 +781,18 @@ typedef struct
   int flags;
 } STATE;
 
+/* used by enter.c */
+
+typedef struct
+{
+  wchar_t *wbuf;
+  size_t wbuflen;
+  size_t lastchar;
+  size_t curpos;
+  size_t begin;
+  int	 tabs;
+} ENTER_STATE;
+
 /* flags for the STATE struct */
 #define M_DISPLAY	(1<<0) /* output is displayed to the user */
 
@@ -704,15 +807,19 @@ typedef struct
 #define M_PENDINGPREFIX (1<<2) /* prefix to write, but character must follow */
 #define M_WEED          (1<<3) /* weed headers even when not in display mode */
 #define M_CHARCONV	(1<<4) /* Do character set conversions */
+#define M_PRINTING	(1<<5) /* are we printing? - M_DISPLAY "light" */
 
 #define state_set_prefix(s) ((s)->flags |= M_PENDINGPREFIX)
 #define state_reset_prefix(s) ((s)->flags &= ~M_PENDINGPREFIX)
 #define state_puts(x,y) fputs(x,(y)->fpout)
 #define state_putc(x,y) fputc(x,(y)->fpout)
 
-void state_prefix_putc(char, STATE *);
+void state_mark_attach (STATE *);
+void state_attach_puts (const char *, STATE *);
+void state_prefix_putc (char, STATE *);
 int  state_printf(STATE *, const char *, ...);
 
+#include "ascii.h"
 #include "protos.h"
 #include "lib.h"
 #include "globals.h"
