@@ -24,7 +24,7 @@
 #include "mailbox.h"
 #include "copy.h"
 #include "keymap.h"
-
+#include "url.h"
 
 #ifdef HAVE_PGP
 #include "pgp.h"
@@ -34,6 +34,10 @@
 #include "imap.h"
 #endif
 
+#ifdef USE_POP
+#include "pop.h"
+#endif
+
 #ifdef BUFFY_SIZE
 #include "buffy.h"
 #endif
@@ -313,11 +317,39 @@ void mx_unlink_empty (const char *path)
 
 int mx_is_imap(const char *p)
 {
-  return p && (*p == '{');
+  url_scheme_t scheme;
+
+  if (!p)
+    return 0;
+
+  if (*p == '{')
+    return 1;
+
+  scheme = url_check_scheme (p);
+  if (scheme == U_IMAP || scheme == U_IMAPS)
+    return 1;
+
+  return 0;
 }
 
 #endif
 
+#ifdef USE_POP
+int mx_is_pop (const char *p)
+{
+  url_scheme_t scheme;
+
+  if (!p)
+    return 0;
+
+  scheme = url_check_scheme (p);
+  if (scheme == U_POP || scheme == U_POPS)
+    return 1;
+
+  return 0;
+}
+#endif
+
 int mx_get_magic (const char *path)
 {
   struct stat st;
@@ -330,6 +362,11 @@ int mx_get_magic (const char *path)
     return M_IMAP;
 #endif /* USE_IMAP */
 
+#ifdef USE_POP
+  if (mx_is_pop (path))
+    return M_POP;
+#endif /* USE_POP */
+
   if (stat (path, &st) == -1)
   {
     dprint (1, (debugfile, "mx_get_magic(): unable to stat %s: %s (errno %d).\n",
@@ -359,6 +396,10 @@ int mx_get_magic (const char *path)
     if (access (tmp, F_OK) == 0)
       return (M_MH);
 
+    snprintf (tmp, sizeof (tmp), "%s/.mew-cache", path);
+    if (access (tmp, F_OK) == 0)
+      return (M_MH);
+
     /* 
      * ok, this isn't an mh folder, but mh mode can be used to read
      * Usenet news from the spool. ;-) 
@@ -415,15 +456,15 @@ int mx_get_magic (const char *path)
  */
 int mx_set_magic (const char *s)
 {
-  if (mutt_strcasecmp (s, "mbox") == 0)
+  if (ascii_strcasecmp (s, "mbox") == 0)
     DefaultMagic = M_MBOX;
-  else if (mutt_strcasecmp (s, "mmdf") == 0)
+  else if (ascii_strcasecmp (s, "mmdf") == 0)
     DefaultMagic = M_MMDF;
-  else if (mutt_strcasecmp (s, "mh") == 0)
+  else if (ascii_strcasecmp (s, "mh") == 0)
     DefaultMagic = M_MH;
-  else if (mutt_strcasecmp (s, "maildir") == 0)
+  else if (ascii_strcasecmp (s, "maildir") == 0)
     DefaultMagic = M_MAILDIR;
-  else if (mutt_strcasecmp (s, "kendra") == 0)
+  else if (ascii_strcasecmp (s, "kendra") == 0)
     DefaultMagic = M_KENDRA;
   else
     return (-1);
@@ -431,6 +472,19 @@ int mx_set_magic (const char *s)
   return 0;
 }
 
+/* mx_access: Wrapper for access, checks permissions on a given mailbox.
+ *   We may be interested in using ACL-style flags at some point, currently
+ *   we use the normal access() flags. */
+int mx_access (const char* path, int flags)
+{
+#ifdef USE_IMAP
+  if (mx_is_imap (path))
+    return imap_access (path, flags);
+#endif
+
+  return access (path, flags);
+}
+
 static int mx_open_mailbox_append (CONTEXT *ctx)
 {
   struct stat sb;
@@ -588,7 +642,7 @@ CONTEXT *mx_open_mailbox (const char *path, int flags, CONTEXT *pctx)
     {
       mx_fastclose_mailbox (ctx);
       if (!pctx)
-	safe_free ((void **) &ctx);
+	FREE (&ctx);
       return NULL;
     }
     return ctx;
@@ -617,10 +671,6 @@ CONTEXT *mx_open_mailbox (const char *path, int flags, CONTEXT *pctx)
    */
   set_option (OPTFORCEREFRESH);
 
-  /* create hash tables */
-  ctx->id_hash = hash_create (1031);
-  ctx->subj_hash = hash_create (1031);
-
   if (!ctx->quiet)
     mutt_message (_("Reading %s..."), ctx->path);
 
@@ -645,6 +695,12 @@ CONTEXT *mx_open_mailbox (const char *path, int flags, CONTEXT *pctx)
       break;
 #endif /* USE_IMAP */
 
+#ifdef USE_POP
+    case M_POP:
+      rc = pop_open_mailbox (ctx);
+      break;
+#endif /* USE_POP */
+
     default:
       rc = -1;
       break;
@@ -667,7 +723,7 @@ CONTEXT *mx_open_mailbox (const char *path, int flags, CONTEXT *pctx)
   {
     mx_fastclose_mailbox (ctx);
     if (!pctx)
-      safe_free ((void **) &ctx);
+      FREE (&ctx);
   }
 
   unset_option (OPTFORCEREFRESH);
@@ -684,12 +740,15 @@ void mx_fastclose_mailbox (CONTEXT *ctx)
   
 #ifdef USE_IMAP
   if (ctx->magic == M_IMAP)
-    imap_fastclose_mailbox (ctx);
+    imap_close_mailbox (ctx);
 #endif /* USE_IMAP */
-  if (ctx->subj_hash)
-    hash_destroy (&ctx->subj_hash, NULL);
+#ifdef USE_POP
+  if (ctx->magic == M_POP)
+    pop_close_mailbox (ctx);
+#endif /* USE_POP */
   if (ctx->id_hash)
     hash_destroy (&ctx->id_hash, NULL);
+  mutt_clear_threads (ctx);
   for (i = 0; i < ctx->msgcount; i++)
     mutt_free_header (&ctx->hdrs[i]);
   safe_free ((void **) &ctx->hdrs);
@@ -734,6 +793,12 @@ static int sync_mailbox (CONTEXT *ctx, int *index_hint)
       rc = imap_sync_mailbox (ctx, 1, index_hint);
       break;
 #endif /* USE_IMAP */
+
+#ifdef USE_POP
+    case M_POP:
+      rc = pop_sync_mailbox (ctx, index_hint);
+      break;
+#endif /* USE_POP */
   }
 
 #if 0
@@ -781,7 +846,8 @@ int mx_close_mailbox (CONTEXT *ctx, int *index_hint)
 
   for (i = 0; i < ctx->msgcount; i++)
   {
-    if (!ctx->hdrs[i]->deleted && ctx->hdrs[i]->read)
+    if (!ctx->hdrs[i]->deleted && ctx->hdrs[i]->read 
+        && !(ctx->hdrs[i]->flagged && option (OPTKEEPFLAGGED)))
       read_msgs++;
   }
 
@@ -805,17 +871,27 @@ int mx_close_mailbox (CONTEXT *ctx, int *index_hint)
     {
       snprintf (buf, sizeof (buf), _("Move read messages to %s?"), mbox);
       if ((move_messages = query_quadoption (OPT_MOVE, buf)) == -1)
+      {
+	ctx->closing = 0;
 	return (-1);
+      }
     }
   }
 
-  if (ctx->deleted)
+  /* 
+   * There is no point in asking whether or not to purge if we are
+   * just marking messages as "trash".
+   */
+  if (ctx->deleted && !(ctx->magic == M_MAILDIR && option (OPTMAILDIRTRASH)))
   {
     snprintf (buf, sizeof (buf), ctx->deleted == 1
 	     ? _("Purge %d deleted message?") : _("Purge %d deleted messages?"),
 	      ctx->deleted);
     if ((purge = query_quadoption (OPT_DELETE, buf)) < 0)
+    {
+      ctx->closing = 0;
       return (-1);
+    }
   }
 
 #ifdef USE_IMAP
@@ -843,7 +919,8 @@ int mx_close_mailbox (CONTEXT *ctx, int *index_hint)
     {
       /* tag messages for moving, and clear old tags, if any */
       for (i = 0; i < ctx->msgcount; i++)
-	if (ctx->hdrs[i]->read && !ctx->hdrs[i]->deleted)
+	if (ctx->hdrs[i]->read && !ctx->hdrs[i]->deleted
+            && !(ctx->hdrs[i]->flagged && option (OPTKEEPFLAGGED))) 
 	  ctx->hdrs[i]->tagged = 1;
 	else
 	  ctx->hdrs[i]->tagged = 0;
@@ -854,16 +931,23 @@ int mx_close_mailbox (CONTEXT *ctx, int *index_hint)
     if (i == 0) /* success */
       mutt_clear_error ();
     else if (i == -1) /* horrible error, bail */
+    {
+      ctx->closing=0;
       return -1;
+    }
     else /* use regular append-copy mode */
 #endif
     {
       if (mx_open_mailbox (mbox, M_APPEND, &f) == NULL)
+      {
+	ctx->closing = 0;
 	return -1;
+      }
 
       for (i = 0; i < ctx->msgcount; i++)
       {
-	if (ctx->hdrs[i]->read && !ctx->hdrs[i]->deleted)
+	if (ctx->hdrs[i]->read && !ctx->hdrs[i]->deleted
+            && !(ctx->hdrs[i]->flagged && option (OPTKEEPFLAGGED)))
         {
 	  if (mutt_append_message (&f, ctx, ctx->hdrs[i], 0, CH_UPDATE_LEN) == 0)
 	  {
@@ -873,6 +957,7 @@ int mx_close_mailbox (CONTEXT *ctx, int *index_hint)
 	  else
 	  {
 	    mx_close_mailbox (&f, NULL);
+	    ctx->closing = 0;
 	    return -1;
 	  }
 	}
@@ -894,7 +979,10 @@ int mx_close_mailbox (CONTEXT *ctx, int *index_hint)
   if (ctx->magic == M_IMAP)
   {
     if ((check = imap_sync_mailbox (ctx, purge, index_hint)) != 0)
+    {
+      ctx->closing = 0;
       return check;
+    }
   }
   else
 #endif
@@ -909,7 +997,10 @@ int mx_close_mailbox (CONTEXT *ctx, int *index_hint)
     if (ctx->changed || ctx->deleted)
     {
       if ((check = sync_mailbox (ctx, index_hint)) != 0)
+      {
+	ctx->closing = 0;
 	return check;
+      }
     }
   }
 
@@ -949,7 +1040,8 @@ void mx_update_tables(CONTEXT *ctx, int committing)
 #define this_body ctx->hdrs[j]->content
   for (i = 0, j = 0; i < ctx->msgcount; i++)
   {
-    if ((committing && !ctx->hdrs[i]->deleted) || 
+    if ((committing && (!ctx->hdrs[i]->deleted || 
+			(ctx->magic == M_MAILDIR && option (OPTMAILDIRTRASH)))) ||
 	(!committing && ctx->hdrs[i]->active))
     {
       if (i != j)
@@ -966,14 +1058,15 @@ void mx_update_tables(CONTEXT *ctx, int committing)
 	  this_body->hdr_offset;
       }
 
-      if(committing)
+      if (committing)
 	ctx->hdrs[j]->changed = 0;
-      else
+      else if (ctx->hdrs[j]->changed)
+	ctx->changed++;
+      
+      if (!committing || (ctx->magic == M_MAILDIR && option (OPTMAILDIRTRASH)))
       {
 	if (ctx->hdrs[j]->deleted)
 	  ctx->deleted++;
-	if (ctx->hdrs[j]->changed)
-	  ctx->changed++;
       }
 
       if (ctx->hdrs[j]->tagged)
@@ -996,9 +1089,7 @@ void mx_update_tables(CONTEXT *ctx, int committing)
 		      ctx->hdrs[i]->content->offset -
 		      ctx->hdrs[i]->content->hdr_offset);
       /* remove message from the hash tables */
-      if (ctx->hdrs[i]->env->real_subj)
-	hash_delete (ctx->subj_hash, ctx->hdrs[i]->env->real_subj, ctx->hdrs[i], NULL);
-      if (ctx->hdrs[i]->env->message_id)
+      if (ctx->id_hash && ctx->hdrs[i]->env->message_id)
 	hash_delete (ctx->id_hash, ctx->hdrs[i]->env->message_id, ctx->hdrs[i], NULL);
       mutt_free_header (&ctx->hdrs[i]);
     }
@@ -1018,6 +1109,7 @@ int mx_sync_mailbox (CONTEXT *ctx, int *index_hint)
 {
   int rc, i;
   int purge = 1;
+  int msgcount, deleted;
 
   if (ctx->dontwrite)
   {
@@ -1068,6 +1160,11 @@ int mx_sync_mailbox (CONTEXT *ctx, int *index_hint)
     }
   }
 
+  /* really only for IMAP - imap_sync_mailbox results in a call to
+   * mx_update_tables, so ctx->deleted is 0 when it comes back */
+  msgcount = ctx->msgcount;
+  deleted = ctx->deleted;
+
 #ifdef USE_IMAP
   if (ctx->magic == M_IMAP)
     rc = imap_sync_mailbox (ctx, purge, index_hint);
@@ -1078,17 +1175,17 @@ int mx_sync_mailbox (CONTEXT *ctx, int *index_hint)
   {
 #ifdef USE_IMAP
     if (ctx->magic == M_IMAP && !purge)
-      mutt_message (_("%d kept."), ctx->msgcount);
+      mutt_message _("Mailbox checkpointed.");
     else
 #endif
-    mutt_message (_("%d kept, %d deleted."), ctx->msgcount - ctx->deleted,
-      ctx->deleted);
-
-    sleep (1); /* allow the user time to read the message */
+    mutt_message (_("%d kept, %d deleted."), msgcount - deleted,
+      deleted);
 
+    mutt_sleep (0);
+    
     if (ctx->msgcount == ctx->deleted &&
 	(ctx->magic == M_MBOX || ctx->magic == M_MMDF) &&
-	!mutt_is_spool(ctx->path) && !option (OPTSAVEEMPTY))
+	!mutt_is_spool (ctx->path) && !option (OPTSAVEEMPTY))
     {
       unlink (ctx->path);
       mx_fastclose_mailbox (ctx);
@@ -1096,10 +1193,22 @@ int mx_sync_mailbox (CONTEXT *ctx, int *index_hint)
     }
 
     /* if we haven't deleted any messages, we don't need to resort */
-    if (purge)
+    /* ... except for certain folder formats which need "unsorted" 
+     * sort order in order to synchronize folders.
+     * 
+     * MH and maildir are safe.  mbox-style seems to need re-sorting,
+     * at least with the new threading code.
+     */
+    if (purge || (ctx->magic != M_MAILDIR && ctx->magic != M_MH))
     {
-      mx_update_tables(ctx, 1);
-      mutt_sort_headers (ctx, 1); /* rethread from scratch */
+#ifdef USE_IMAP
+      /* IMAP does this automatically after handling EXPUNGE */
+      if (ctx->magic != M_IMAP)
+#endif
+      {
+	mx_update_tables (ctx, 1);
+	mutt_sort_headers (ctx, 1); /* rethread from scratch */
+      }
     }
   }
 
@@ -1167,6 +1276,13 @@ MESSAGE *mx_open_new_message (CONTEXT *dest, HEADER *hdr, int flags)
   msg->magic = dest->magic;
   msg->write = 1;
 
+  if (hdr)
+  {
+    msg->flags.flagged = hdr->flagged;
+    msg->flags.replied = hdr->replied;
+    msg->flags.read    = hdr->read;
+  }
+  
   if (func (msg, dest, hdr) == 0)
   {
     if (dest->magic == M_MMDF)
@@ -1243,6 +1359,11 @@ int mx_check_mailbox (CONTEXT *ctx, int *index_hint, int lock)
       case M_IMAP:
 	return (imap_check_mailbox (ctx, index_hint));
 #endif /* USE_IMAP */
+
+#ifdef USE_POP
+      case M_POP:
+	return (pop_check_mailbox (ctx, index_hint));
+#endif /* USE_POP */
     }
   }
 
@@ -1270,7 +1391,12 @@ MESSAGE *mx_open_message (CONTEXT *ctx, int msgno)
       char path[_POSIX_PATH_MAX];
       
       snprintf (path, sizeof (path), "%s/%s", ctx->path, cur->path);
-      if ((msg->fp = fopen (path, "r")) == NULL)
+      
+      if ((msg->fp = fopen (path, "r")) == NULL && errno == ENOENT &&
+	  ctx->magic == M_MAILDIR)
+	msg->fp = maildir_open_find_message (ctx->path, cur->path);
+      
+      if (msg->fp == NULL)
       {
 	mutt_perror (path);
 	dprint (1, (debugfile, "mx_open_message: fopen: %s: %s (errno %d).\n",
@@ -1289,6 +1415,15 @@ MESSAGE *mx_open_message (CONTEXT *ctx, int msgno)
     }
 #endif /* USE_IMAP */
 
+#ifdef USE_POP
+    case M_POP:
+    {
+      if (pop_fetch_message (msg, ctx, msgno) != 0)
+	FREE (&msg);
+      break;
+    }
+#endif /* USE_POP */
+
     default:
       dprint (1, (debugfile, "mx_open_message(): function not implemented for mailbox type %d.\n", ctx->magic));
       FREE (&msg);
@@ -1356,7 +1491,7 @@ int mx_commit_message (MESSAGE *msg, CONTEXT *ctx)
   }
   
   if (r == 0 && (ctx->magic == M_MBOX || ctx->magic == M_MMDF || ctx->magic == M_KENDRA)
-      && fflush (msg->fp) == EOF)
+      && (fflush (msg->fp) == EOF || fsync (fileno (msg->fp)) == -1))
   {
     mutt_perror _("Can't write message");
     r = -1;
@@ -1373,6 +1508,9 @@ int mx_close_message (MESSAGE **msg)
   if ((*msg)->magic == M_MH || (*msg)->magic == M_MAILDIR
 #ifdef USE_IMAP
       || (*msg)->magic == M_IMAP
+#endif
+#ifdef USE_POP
+      || (*msg)->magic == M_POP
 #endif
       )
   {
@@ -1417,60 +1555,67 @@ void mx_alloc_memory (CONTEXT *ctx)
 /* this routine is called to update the counts in the context structure for
  * the last message header parsed.
  */
-void mx_update_context (CONTEXT *ctx)
+void mx_update_context (CONTEXT *ctx, int new_messages)
 {
-  HEADER *h = ctx->hdrs[ctx->msgcount];
+  HEADER *h;
+  int msgno;
 
+  for (msgno = ctx->msgcount - new_messages; msgno < ctx->msgcount; msgno++)
+  {
+    h = ctx->hdrs[msgno];
 
 
 
 #ifdef HAVE_PGP
-  /* NOTE: this _must_ be done before the check for mailcap! */
-  h->pgp = pgp_query (h->content);
+    /* NOTE: this _must_ be done before the check for mailcap! */
+    h->pgp = pgp_query (h->content);
 #endif /* HAVE_PGP */
 
-  if (!ctx->pattern)
-  {
-    ctx->v2r[ctx->vcount] = ctx->msgcount;
-    h->virtual = ctx->vcount++;
-  }
-  else
-    h->virtual = -1;
-  h->msgno = ctx->msgcount;
-  ctx->msgcount++;
-
-  if (h->env->supersedes)
-  {
-    HEADER *h2 = hash_find (ctx->id_hash, h->env->supersedes);
-
-    /* safe_free (&h->env->supersedes); should I ? */
-    if (h2)
+    if (!ctx->pattern)
     {
-      h2->superseded = 1;
-      if (option (OPTSCORE)) 
-	mutt_score_message (ctx, h2, 1);
+      ctx->v2r[ctx->vcount] = msgno;
+      h->virtual = ctx->vcount++;
+    }
+    else
+      h->virtual = -1;
+    h->msgno = msgno;
+
+    if (h->env->supersedes)
+    {
+      HEADER *h2;
+
+      if (!ctx->id_hash)	
+	ctx->id_hash = mutt_make_id_hash (ctx);
+
+      h2 = hash_find (ctx->id_hash, h->env->supersedes);
+
+      /* safe_free (&h->env->supersedes); should I ? */
+      if (h2)
+      {
+	h2->superseded = 1;
+	if (option (OPTSCORE)) 
+	  mutt_score_message (ctx, h2, 1);
+      }
+    }
+
+    /* add this message to the hash tables */
+    if (ctx->id_hash && h->env->message_id)
+      hash_insert (ctx->id_hash, h->env->message_id, h, 0);
+
+    if (option (OPTSCORE)) 
+      mutt_score_message (ctx, h, 0);
+
+    if (h->changed)
+      ctx->changed = 1;
+    if (h->flagged)
+      ctx->flagged++;
+    if (h->deleted)
+      ctx->deleted++;
+    if (!h->read)
+    {
+      ctx->unread++;
+      if (!h->old)
+	ctx->new++;
     }
   }
-
-  /* add this message to the hash tables */
-  if (h->env->message_id)
-    hash_insert (ctx->id_hash, h->env->message_id, h, 0);
-  if (h->env->real_subj)
-    hash_insert (ctx->subj_hash, h->env->real_subj, h, 1);
-
-  if (option (OPTSCORE)) 
-    mutt_score_message (ctx, h, 0);
-  
-  if (h->changed)
-    ctx->changed = 1;
-  if (h->flagged)
-    ctx->flagged++;
-  if (h->deleted)
-    ctx->deleted++;
-  if (!h->read)
-  {
-    ctx->unread++;
-    if (!h->old)
-      ctx->new++;
-  }
 }
