@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1996-8 Michael R. Elkins <me@cs.hmc.edu>
+ * Copyright (C) 1996-2000 Michael R. Elkins <me@cs.hmc.edu>
  * 
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -13,7 +13,7 @@
  * 
  *     You should have received a copy of the GNU General Public License
  *     along with this program; if not, write to the Free Software
- *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
  */ 
 
 #include "mutt.h"
@@ -33,7 +33,9 @@ static sort_t *AuxSort = NULL;
   set_option(OPTAUXSORT); \
   code = AuxSort(a,b); \
   unset_option(OPTAUXSORT); \
-}
+} \
+if (!code) \
+  code = (*((HEADER **)a))->index - (*((HEADER **)b))->index;
 
 int compare_score (const void *a, const void *b)
 {
@@ -176,8 +178,10 @@ void mutt_sort_headers (CONTEXT *ctx, int init)
 {
   int i;
   HEADER *h;
+  THREAD *thread, *top;
   sort_t *sortfunc;
   
+  init = 1; /* XXX temporary bug workaround (hopefully!) */
   unset_option (OPTNEEDRESORT);
 
   if (!ctx)
@@ -197,16 +201,21 @@ void mutt_sort_headers (CONTEXT *ctx, int init)
   if (!ctx->quiet)
     mutt_message _("Sorting mailbox...");
 
-  /* threads may be bogus, so clear the links */
-  if (init)
-    mutt_clear_threads (ctx);
-
-  if (option (OPTNEEDRESCORE))
+  if (option (OPTNEEDRESCORE) && option (OPTSCORE))
   {
     for (i = 0; i < ctx->msgcount; i++)
-      mutt_score_message (ctx->hdrs[i]);
-    unset_option (OPTNEEDRESCORE);
+      mutt_score_message (ctx, ctx->hdrs[i], 1);
   }
+  unset_option (OPTNEEDRESCORE);
+
+  if (option (OPTRESORTINIT))
+  {
+    unset_option (OPTRESORTINIT);
+    init = 1;
+  }
+
+  if (init && ctx->tree)
+    mutt_clear_threads (ctx);
 
   if ((Sort & SORT_MASK) == SORT_THREADS)
   {
@@ -215,7 +224,11 @@ void mutt_sort_headers (CONTEXT *ctx, int init)
        subthreads need to be resorted */
     if (option (OPTSORTSUBTHREADS))
     {
-      ctx->tree = mutt_sort_subthreads (ctx->tree, mutt_get_sort_func (SortAux));
+      i = Sort;
+      Sort = SortAux;
+      if (ctx->tree)
+	ctx->tree = mutt_sort_subthreads (ctx->tree, 1);
+      Sort = i;
       unset_option (OPTSORTSUBTHREADS);
     }
     mutt_sort_threads (ctx, init);
@@ -224,17 +237,12 @@ void mutt_sort_headers (CONTEXT *ctx, int init)
 	   (AuxSort = mutt_get_sort_func (SortAux)) == NULL)
   {
     mutt_error _("Could not find sorting function! [report this bug]");
-    sleep (1);
+    mutt_sleep (1);
     return;
   }
   else 
     qsort ((void *) ctx->hdrs, ctx->msgcount, sizeof (HEADER *), sortfunc);
 
-  /* the threading function find_reference() needs to know how the mailbox
-   * is currently sorted in memory in order to speed things up a bit
-   */
-  ctx->revsort = (Sort & SORT_REVERSE) ? 1 : 0;
-
   /* adjust the virtual message numbers */
   ctx->vcount = 0;
   for (i = 0; i < ctx->msgcount; i++)
@@ -252,12 +260,16 @@ void mutt_sort_headers (CONTEXT *ctx, int init)
   /* re-collapse threads marked as collapsed */
   if ((Sort & SORT_MASK) == SORT_THREADS)
   {
-    h = ctx->tree;
-    while (h)
+    top = ctx->tree;
+    while ((thread = top) != NULL)
     {
+      while (!thread->message)
+	thread = thread->child;
+      h = thread->message;
+
       if (h->collapsed)
 	mutt_collapse_thread (ctx, h);
-      h = h->next;
+      top = top->next;
     }
     mutt_set_virtual (ctx);
   }
